/*
Copyright (c) 1991-1995 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: call.c,v 1.220 1997/09/18 05:53:49 spreitze Exp $ */
/* Last edited by Mike Spreitzer September 17, 1997 10:42 pm PDT */

#define _POSIX_SOURCE

#include <stddef.h>

#include "iluntrnl.h"
#include "protocol.h"
#include "connect.h"
#include "transprt.h"
#include "port.h"
#include "call.h"
#include "object.h"
#include "method.h"
#include "server.h"
#include "type.h"

/*L1, L2 unconstrained*/

const char     *const ilu_PENames[ilu_ProtocolException_Not + 1] = {
  "Success",
  "NoSuchClassAtServer",
  "ClassVersionMismatch",
  "NoSuchMethodOnClass",
  "GarbageArguments",
  "Unknown",
  "LostConnection",
  "RequestRejected",
  "RequestTimeout",
  "Not"
};

ilu_cardinal _ilu_SafeStrlen(ilu_string s)
{
  if (s == NIL)
    return 0;
  else
    return (strlen(s));
}

ilu_Class ilu_IntroTypeOfCall (ilu_Call call)
{
  if (call == NIL)
    return (NIL);
  else
    return (call->ca_intro_type);
}

ilu_Connection ilu_ConnectionOfCall (ilu_Call call)
{
  if (call == NIL)
    return (NIL);
  else
    return (call->ca_connection);
}

ilu_boolean ilu_CallNeedsSizing (ilu_Call call)
{
  if (call == NIL)
    return (ilu_FALSE);
  else
    return (protocol_needs_sizing(connection_protocol(call_connection(call))));
}

ilu_Method ilu_MethodOfCall (ilu_Call call)
{
  if (call == NIL)
    return (NIL);
  else
    return (call->ca_method);
}

static void
BuildCall(ilu_Call call, ilu_Connection connection, ilu_Server server,
	  ilu_cardinal serialNumber, ilu_boolean incoming)
{
  static ilu_Call_s nocall = {0};
  *call = nocall;
  call->ca_SN = serialNumber;
  call->ca_server = server;
  call->ca_connection = connection;
  call->ca_pl = NIL;
  call->ca_si = NIL;
  call->ca_incoming = incoming;
  call->ca_ios = ilu_ciosNone;
  call->ca_ms = ilu_cmsNo;
  call->ca_pe = ilu_ProtocolException_Success;
  call->ca_tryIndex = 0;
  return;
}

/*
 * Returns TRUE iff connection should continue to be monitored.
 * Returns FALSE if (but not only if) raising an ilu_Error.
 * Caller pins (conn) into existence.
 */
/*L1 = {cmu, server}; Main Remnant; L2 >= {conn's callmu, iomu, waitmu}*/
static          ilu_boolean
ProcessExtraInput(ilu_Connection conn, ilu_Call dummy,
		  ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Server      server = connection_server(conn);
  ilu_Protocol    proto = connection_protocol(conn);
  ilu_ReadHeaderResultCode rhrc;
  ilu_PacketType  packetType;
  ilu_cardinal    packetSN;
  if (!_ilu_ExitServerMutex(server, TRUE, err))
    return FALSE;
  if (!ilu_ExitMutex(ilu_cmu, TRUE, err))
    return FALSE;
  /* Main Invariant; Call-IHi(dummy) */
  rhrc = protocol_read_header(proto, dummy,
			      &packetType, &packetSN, err);
  switch (rhrc) {
  case ilu_rhrc_ok:
    /*
     * We know no calls are in progess, so nobody cares about this
     * reply, so flush it.  When there exist other kinds of
     * messages, this code will have to become prepared to handle
     * them.
     */
    if (!protocol_discard_input(proto, dummy, err))
      goto closeout;
  case ilu_rhrc_nothing:
  case ilu_rhrc_handled:
    if (!ilu_ReEnterMutex(ilu_cmu, err))
      return FALSE;
    if (!_ilu_EnterServerMutex(server, TRUE, err))
      return FALSE;
    return TRUE;
  case ilu_rhrc_eof:
  case ilu_rhrc_error:
closeout:
    if (ILU_ERROK(*err)) {
      ILU_NOTE(CONNECTION_DEBUG,
	    ("Server closed connection %p to <%s>.\n",
	     conn, server->sr_id));
    } else {
      ILU_NOTE(CONNECTION_DEBUG,
	    ("Closing connection %p to <%s>, due to error %s"
	     " (raised at %s:%d).\n",
	     conn, server->sr_id, ILU_ERR_NAME(*err),
	     ilu_ErrorFile(err), ilu_ErrorLine(err)));
    }
    goto dunX;
  default:
    if (!ilu_Check(0, err))
      goto dunX;
    return TRUE;
  }
dunX:
  if (!ilu_ReEnterMutex(ilu_cmu, err))
    return FALSE;
  if (!_ilu_EnterServerMutex(server, TRUE, err))
    return FALSE;
  _ilu_CloseIoingConnection(conn, ILU_ERRNOK(*err));
  return FALSE;
}

/*Main Invariant holds; L2 disjoint {conn's callmu, iomu, waitmu}*/
static void ReadExtraMsg(ilu_refany rock)
{
  ilu_Connection  conn = (ilu_Connection) rock;
  ilu_Protocol    proto = connection_protocol(conn);
  ilu_Server      server = connection_server(conn);
  ilu_boolean     prco = protocol_concurrent(proto);
  ilu_Call_s      dummyCall;
  ilu_Error       lerr;
  ilu_boolean     rereg = FALSE, initted = FALSE;
  if (!ilu_EnterMutex(ilu_cmu, &lerr))
    goto dun0;
  if (!_ilu_EnterServerMutex(server, FALSE, &lerr))
    goto dun1;
  if (conn->co_mucall != NIL)
    /* Spurrious call; stop here. */
    goto dun2;
  if (conn->co_nOuts != 0)
    goto dun2;
  if (conn->co_waiting)
    goto dun2;
  if (connection_closed(conn))
    goto dun2;
  if (!ilu_ClearConnectionInputHandler(conn, &lerr))
    goto dun2;
  BuildCall(&dummyCall, conn, server, 0, FALSE);
  dummyCall.ca_disownWait = FALSE;
  if (!(*proto->pr_init_call) (&dummyCall, &lerr))
    goto dun2;
  initted = TRUE;
  _ilu_TakeConnWait(conn, &dummyCall);
  if (!_ilu_EnterConnCall(conn, &dummyCall, TRUE, &lerr))
    goto dun3;
  if (!_ilu_EnterConnIO(conn, TRUE, &lerr))
    goto dun4;
  /* L1 = {cmu, server}; L2 >= {conn's callmu, iomu, waitmu} */
  if (connection_closed(conn))
    goto dun5;
  rereg = ProcessExtraInput(conn, &dummyCall, &lerr);
dun5:
  if (ILU_ERROK(lerr) && proto->pr_prefinish_call != NULLFN) {
    (void) (*proto->pr_prefinish_call) (&dummyCall, &lerr);
    rereg = rereg && !connection_closed(conn);
  }
  (void) _ilu_ReleaseConnIO(conn, TRUE, &lerr);
dun4:
  (void) _ilu_ReleaseConnCall(conn, &dummyCall, TRUE, &lerr);
dun3:
  (void) _ilu_ReleaseConnWait(conn, &dummyCall, TRUE, &lerr);
dun2:
  if (!rereg)
    (void) _ilu_MaybeFreeConnection(conn);
  (void) _ilu_ExitServerMutex(server, TRUE, &lerr);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, &lerr);
  if (rereg) {
    if (!ilu_SetConnectionInputHandler(conn, ReadExtraMsg, conn, &lerr))
      goto dun0;
  }
dun0:
  ILU_HANDLED(lerr);
  if (initted) {
    (*proto->pr_finish_call) (&dummyCall, &lerr);
    ILU_HANDLED(lerr);
  }
  return;
}

/**Main Remnant;
   Before: L1 >= {cmu, s}, (L1-{s}).sup = cmu;
   After:  res>0 <=> L1 >= {cmu}, res>1 <=> L1 >= {s} */
static int
OpenConn(ilu_string pinfo, ilu_Protocol proto,
	 ilu_TransportInfo tinfo, ilu_TransportCreator tcr,
	 ilu_Server s, ilu_Passport pp,
	 ilu_Connection * pconn, ilu_Connection * new_conn,
	 ILU_ERRS((IoErrs, bad_locks,
		   inv_objref, no_resources)) * err)
{
  ilu_Connection  conn = NIL;
  ilu_integer     dfd, cdfd;
  ilu_Transport   t;
  *pconn = NIL;
  _ilu_DHoldsTCR(tcr, 1);
  if (!_ilu_ExitServerMutex(s, TRUE, err))
    goto dun1H;
  dfd = (*tcr->tcr_dfd) (tcr, FALSE);
  if ((ilu_fdbudget < ilu_fdstaken + dfd)
      && (dfd > 0)) {
    ILU_ERRS((internal, WrongLocks)) lerr;
    lerr = _ilu_ReduceFdsTo(ilu_fdbudget - dfd);
    ILU_ERR_SWITCH(lerr) {
      ILU_SUCCESS_CASE;
      ILU_ERR_CASE(WrongLocks, v) {
	ILU_ERR_CONS0(bad_locks, err, 0);
	goto dun1H;
      }
    } ILU_ERR_ENDSWITCH;
    if ((ilu_fdbudget < ilu_fdstaken + dfd)
	&& (dfd > 0)) {
      ILU_NOTE(CALL_DEBUG,
	       ("call.c:OpenConn: FD budget exhausted.\n"));
      ILU_ERR_CONS1(no_resources, err, minor, ilu_nrm_fds, 0);
      goto dun1H;
    }
  }
  ilu_DeltaFD(dfd);
  if (!ilu_ExitMutex(ilu_cmu, TRUE, err))
    return 0;
  t = (*tcr->tcr_createTransport) (tcr, TRUE, &cdfd, pp, err);
  if (!ilu_ReEnterMutex(ilu_cmu, err))
    return 0;
  _ilu_DHoldsTCR(tcr, -1);
  ilu_DeltaFD(cdfd - dfd);
  if (t == NIL) {
    if (_ilu_EnterServerMutex(s, FALSE, err)) {
      if (!server_is_true(s) && _ilu_CompareTinfo(tinfo, s->sr_tinfo))
	s->sr_cfails = ilu_TRUE;
      ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_ts, 0);
      (void) _ilu_ExitServerMutex(s, TRUE, err);
    } else
      ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_ts, 0);
    return 1;
  }
  if (!_ilu_EnterServerMutex(s, FALSE, err))
    return 1;
  conn = _ilu_CreateConnection(t, tinfo, NIL, proto, pinfo, NIL, s,
			       pp, err);
  if (conn == NIL)
    return 2;
  *pconn = conn;
  if (ilu_CanCondition())
    *new_conn = conn;
  else {
    conn->co_lsrCares = FALSE;
  }
  return 3;
dun1H:
  _ilu_DHoldsTCR(tcr, -1);
  return 1;
}

/**Main Invariant holds;
   after: success => Call-Locking(call, OHi)*/

ilu_boolean
ilu_StartCall(ilu_Call_s * c, ilu_Server s, ilu_Class intro_type,
	      ilu_Method method, ilu_LanguageIndex lang,
	      ilu_Passport pp,
	      ilu_Connection * new_conn,
	      ILU_ERRS((IoErrs, barrier, bad_locks,
			inv_objref, no_resources)) * err)
{
  return ilu_FullStartCall(c, s, intro_type, method, lang, pp, NIL,
			   NIL, new_conn, err);
}

ilu_boolean
ilu_FullStartCall(ilu_Call_s * c, ilu_Server s, ilu_Class intro_type,
		  ilu_Method method, ilu_LanguageIndex lang,
		  ilu_Passport pp,
		  ilu_Serializer si,	/* OPTIONAL */
		  ilu_Pipeline pl,	/* OPTIONAL */
		  ilu_Connection * new_conn,
		  ILU_ERRS((IoErrs, barrier, bad_locks,
			    inv_objref, no_resources)) * err)
{
  ilu_Connection  conn = NIL;
  ilu_integer     dfd, cdfd;
  ilu_TransportInfo tinfo;
  ilu_string      pinfo;
  ilu_Protocol    proto;
  ilu_TransportCreator tcr = ILU_NIL;
  ilu_Transport   t;
  ilu_boolean     isnew = FALSE;
  ilu_Error       lerr = ILU_INIT_NO_ERR;
  int             openres;

  ILU_CLER(*err);
  *new_conn = NIL;
  if (s == NIL || pl && !pl->pl_lsrCares || si && !si->si_lsrCares)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  if (si && si->si_server != s)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_serialVsServer,
			 FALSE);

  if ((server_is_true(s) &&
       (method_id(method) > 0xFF80))
  /* calling "internal" method on true server */
      || (server_true_language(s) == lang)
  /* calling through kernel for should-be direct call */
    )
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_true, FALSE);

  if (method->me_asynchronous &&
      !(server_is_true(s) || s->sr_tcr->tcr_reliable))
    return ILU_ERR_CONS1(bad_param, err, minor,
			 ilu_bpm_asynch_unreliable, FALSE);

  BuildCall(c, NIL, s, 0, FALSE);
  c->ca_intro_type = intro_type;
  c->ca_method = method;
  if (!ilu_EnterMutex(ilu_cmu, err))
    goto dun0;
  if (!_ilu_EnterServerMutex(s, FALSE, err))
    goto dun1;
  /* L1 = {cmu, s}; ILU_ERROK(*err) */
  if (si) {
    if (si->si_busy) {
      ILU_ERR_CONS1(bad_param, err, minor,
		    ilu_bpm_serialConcurrent,
		    (void) 6);
      goto dun2x;
    }
    si->si_busy = TRUE;
    c->ca_si = si;
  }
  if (si && si->si_conn) {
    conn = si->si_conn;
    if (!ilu_Check(conn->co_serialer == si && !conn->co_mucall, err))
      goto dun2;
    if (conn->co_closed || conn->co_doomed || conn->co_closing) {
      ilu_boolean     isBarrier = conn->co_lastWorking;
      conn->co_serialer = NIL;
      si->si_conn = NIL;
      ILU_NOTE(CONNECTION_DEBUG,
	       ("ilu_StartCall: dissociating serializer %p"
		" and connection %p.\n",
		si, conn));
      if (isBarrier) {
	ILU_NOTE(CALL_DEBUG,
		 ("ilu_StartCall(serializer %p, connection %p) is"
		  " barrier because conn. closed or closing.\n",
		  si, conn));
	ILU_ERR_CONS0(barrier, err, (void) 6);
      }
      _ilu_MaybeFreeConnection(conn);
      conn = NIL;
      if (isBarrier)
	goto dun2;
    }
  }
  if (!conn) {
    for (conn = server_connections(s);
	 (conn != NIL)
	 && (conn->co_mucall || conn->co_serialer ||
	     conn->co_doomed || conn->co_closing ||
	     (connection_concurrent(conn) ? si != NIL
	      : (conn->co_pipeline && conn->co_pipeline != pl)));
	 conn = connection_next(conn))
      0;
  }
  if (conn == NIL) {
    /* conn==NIL, L1={cmu, s}, L2 as at entry */
    if (server_is_true(s)) {
      tinfo = port_tinfo(server_local_port(s));
      pinfo = port_pinfo(server_local_port(s));
      proto = port_protocol(server_local_port(s));
      tcr = port_transport_creator(server_local_port(s));
    } else {
      tinfo = s->sr_tinfo;
      tcr = s->sr_tcr;
      pinfo = s->sr_pinfo;
      proto = s->sr_protocol;
    }
    if (si && protocol_concurrent(proto)) {
      ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_conc_serial,
		    (void) 6);
      goto dun2;
    }
    if (si && !tcr->tcr_reliable) {
      ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_serialVsTransport,
		    (void) 6);
      goto dun2;
    }
    openres = OpenConn(pinfo, proto, tinfo, tcr, s, pp,
		       &conn, new_conn, err);
    switch (openres) {
    case 3:
      break;
    case 2:
      goto dun2;
    case 1:
      goto dun1;
    case 0:
      goto dun0;
    }
    isnew = TRUE;
  }
  /* conn != NIL && conn's IO and call mutexes not held */
  if (!ilu_Check(!connection_closed(conn), err))
    goto dun2;
  if (!ilu_Check(!connection_incoming(conn), err))
    goto dun2;
  if (!ilu_Check((conn->co_nOuts == 0 || conn->co_pipeline
		  || connection_concurrent(conn)),
		 err))
    goto dun2;
  if (!ilu_Check(conn->co_pipeline == pl || !conn->co_pipeline, err))
    goto dun2;
  if (si && !si->si_conn) {
    si->si_conn = conn;
    conn->co_serialer = si;
    ILU_NOTE(CONNECTION_DEBUG,
	     ("ilu_StartCall: associating serializer %p"
	      "and connection %p.\n",
	      si, conn));
  }
  if (pl && !connection_concurrent(conn)) {
    pl->pl_nCalls++;
    c->ca_pl = pl;
    conn->co_pipeline = pl;
  }
  c->ca_disownWait = conn->co_waiting && !_ilu_CanCondition();
  if (!_ilu_CanCondition() && conn->co_nOuts == 0 && !isnew) {
    if (!ilu_Check(!conn->co_waiting, err))
      goto dun2;
    if (!ilu_ClearConnectionInputHandler(conn, err))
      goto dun2;
  }
  if (!_ilu_EnterConnCall(conn, c, FALSE, err))
    goto dun2;
  if (!_ilu_EnterConnIO(conn, FALSE, err))
    goto dun3;
  if (!ilu_Check(!connection_closish(conn), err))
    goto dun4;
  c->ca_ms = ilu_cmsHi;
  c->ca_msInput = FALSE;
  proto = conn->co_protocol;
  /* conn != NIL && conn's IO and call mutexes held */
  if (conn->co_next_sn > 0xFFFFFF)	/* why not 0xFFFF ? */
    conn->co_next_sn = 1;
  c->ca_connection = conn;
  c->ca_SN = conn->co_next_sn++;
  conn->co_nCalls += 1;
  if (!(*proto->pr_init_call) (c, err)) {
    conn->co_nCalls -= 1;
    goto dun2;
  }
  goto dun2x;
dun4:
  (void) _ilu_ReleaseConnIO(conn, TRUE, err);
dun3:
  (void) _ilu_ReleaseConnCall(conn, c, TRUE, err);
dun2:
  if (si)
    si->si_busy = FALSE;
  if (pl && c->ca_pl == pl) {
    pl->pl_nCalls -= 1;
  }
dun2x:
  if (conn && conn->co_nCalls == 0)
    conn->co_pipeline = NIL;
  _ilu_MaybeFreePipeline(pl);
  _ilu_MaybeFreeSerializer(si);
  if (conn)
    _ilu_MaybeFreeConnection(conn);
  (void) _ilu_ExitServerMutex(s, TRUE, err);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
dun0:
  if (ILU_ERROK(*err)) {
    ILU_NOTE(CALL_DEBUG,
	     ("%-20.20s(%p over %p to \"%s\" #%lu,"
	      " %s.%s, pl=%p, si=%p).\n",
	      "ilu_StartCall", c, call_connection(c),
	      call_server_id(c), call_serial_number(c),
	      class_name(intro_type), method_name(method),
	      pl, si));
    c->ca_caller = pp;
    return TRUE;
  } else {
    ILU_NOTE(CALL_DEBUG,
	     ("%-20.20s(to %s, %s.%s, pl=%p, si=%p"
	      ") raises err %s (from %s ln %d).\n",
	      "ilu_StartCall", server_id(s),
	      class_name(intro_type), method_name(method),
	      pl, si, ILU_ERR_NAME(*err),
	      ilu_ErrorFile(err), ilu_ErrorLine(err)));
    ILU_HANDLED(lerr);
    return FALSE;
  }
}

/**Before: Call-Invariant(call, err).
   After:  Main Invariant holds,
	   (L2 disjoint {call's conn's callmu, iomu}), and
	   (L2 disjoint {call's conn's waitmu} iff !call->ca_disownWait)
	   if possible (*err indicates bad_locks or broken_locks
	   when not possible).*/
void
ilu_FinishCall(ilu_Call call, ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Server      s;
  ilu_Connection  conn;
  ilu_Protocol    proto;
  ILU_ERRS((bad_locks, broken_locks, internal)) lerr = ILU_INIT_NO_ERR;
  ilu_boolean     closeconn = FALSE, rewatch = FALSE;
  ilu_Serializer  si = NIL;

#ifdef ENABLE_DEBUGGING
  if (call->ca_incoming) {
    ILU_NOTE(CALL_DEBUG,
	  ("%-20.20s(%p from %s #%lu) => err=%s ca_pe=%s\n",
	   "ilu_FinishCall", call, call_connection_id(call),
	   (unsigned long) call_serial_number(call),
	   ILU_ERR_NAME(*err), ilu_PEName(call->ca_pe)
	   ));
  } else {
    ILU_NOTE(CALL_DEBUG,
	     ("%-20.20s(%p over %p to \"%s\" #%lu, pl=%p, si=%p"
	      ") => err=%s ca_pe=%s\n",
	      "ilu_FinishCall", call, call_connection(call),
	      call_server_id(call),
	      (unsigned long) call_serial_number(call),
	      call->ca_pl, call->ca_si,
	      ILU_ERR_NAME(*err), ilu_PEName(call->ca_pe)
	      ));
  }
#endif				/* ENABLE_DEBUGGING */
  ILU_ERR_SWITCH(*err) {
    ILU_SUCCESS_CASE
      0 /* statement with no effect */ ;
    ILU_ERR_CASE(comm_failure, v)
      closeconn = TRUE;
    ILU_ERR_CASE2(bad_locks, broken_locks)
    /*
     * Here we leak like a sieve, but we can't free everything
     * reliably, and so we leave everything lying around, in hopes
     * that it's useful for debugging.
     */
      return;
    ILU_ERR_ELSE
      0 /* statement with no effect */ ;
  } ILU_ERR_ENDSWITCH;
  s = call_server(call);
  conn = call_connection(call);
  proto = conn ? connection_protocol(conn) : NIL;
  _ilu_Assert((call_incoming(call) ||
	       call->ca_pe == ilu_ProtocolException_Success),
	      "client PE");
  if (!ilu_EnterMutex(ilu_cmu, &lerr))
    goto dun0;
  if (!_ilu_EnterServerMutex(s, FALSE, &lerr))
    goto dun1;
  if (call->ca_ms != ilu_cmsHi)
    call->ca_msInput = FALSE;
  /* L2 >= {waitmu} iff ca_disownWait || ca_msInput && !ca_dontWait */
  if (call->ca_ms == ilu_cmsNo && !closeconn
      && !(conn && proto->pr_prefinish_call)
      && ((call->ca_pe == ilu_ProtocolException_Success
	   && ILU_ERROK(*err))
	  || !call_incoming(call)))
    goto dun3;
  _ilu_Assert(conn != NIL, "FinishCall slow path && !conn");
  if (call->ca_ms != ilu_cmsHi) {
    ilu_boolean     conc = connection_concurrent(conn);
    if (!ilu_Check(call->ca_ios == ilu_ciosNone, &lerr))
      goto dun2;
    if (call->ca_ms == ilu_cmsNo || conc || conn->co_pipeline) {
      if (!_ilu_EnterConnCall(conn, call, FALSE, &lerr))
	goto dun2;
      if ((conc || conn->co_pipeline) && call->ca_ms == ilu_cmsLo) {
	if (!ilu_Check(conn->co_nOuts > 0, &lerr))
	  goto dun4;
	conn->co_nOuts--;
      }
    }
    if (!_ilu_EnterConnIO(conn, FALSE, &lerr))
      goto dun4;
    if (connection_closed(conn))
      closeconn = TRUE;
  }
  /** L2 >= {call's conn's callmu, iomu};
      L2 >= {waitmu} iff ca_disownWait || ca_msInput && !ca_dontWait */
  if (closeconn)
    _ilu_CloseIoingConnection(conn, TRUE);
  else {
    if (call->ca_ios == ilu_ciosNone &&
	(!call_incoming(call) ||
	 call->ca_pe == ilu_ProtocolException_Success && ILU_ERROK(*err)))
      goto dunZ;
    if (!_ilu_ExitServerMutex(s, TRUE, &lerr))
      return;
    if (!ilu_ExitMutex(ilu_cmu, TRUE, &lerr))
      return;
    switch (call->ca_ios) {
    case ilu_ciosIn:
      if (!ilu_Check(call->ca_msInput, &lerr))
	goto dunX;
      if (!proto->pr_discard_input(call, &lerr))
	goto dunX;
      break;
    case ilu_ciosOut:
      if (!proto->pr_discard_output(call, &lerr))
	goto dunX;
    case ilu_ciosNone:
      break;
    default:
      if (!ilu_Check(FALSE, &lerr))
	goto dunX;
    }
    if (call_incoming(call) &&
	(call->ca_pe != ilu_ProtocolException_Success ||
	 ILU_ERRNOK(*err))) {
      ilu_cardinal    exsize;
      if (call->ca_pe == ilu_ProtocolException_Success)
	call->ca_pe = ilu_ProtocolException_Unknown;
      ILU_CLER(*err);
      exsize = ilu_BeginSizingException(call, -(ilu_integer) call->ca_pe,
					err);
      if (ILU_ERRNOK(*err))
	goto dunX;
      ILU_NOTE(CALL_DEBUG,
	    ("%-20.20s(%p %s #%lu, excn ilu_ProtocolException_%s)\n",
	     "pr_begin_exception", call, call_connection_id(call),
	     (unsigned long) call_serial_number(call),
	     ilu_PEName(call->ca_pe)));
      if (!protocol_begin_exception(proto, call, 0, call->ca_pe,
				    exsize, err))
	goto dunX;
      ILU_NOTE(CALL_DEBUG,
	    ("%-20.20s(%p %s #%lu)\n", "pr_finish_exception",
	     call, call_connection_id(call),
	     (unsigned long) call_serial_number(call)));
      if (!protocol_finish_exception(proto, call, err))
	goto dunX;
    }
dunX:
    if (!ilu_ReEnterMutex(ilu_cmu, err))
      return;
    if (!_ilu_EnterServerMutex(s, TRUE, err))
      return;
dunZ:
    if (ILU_ERROK(*err))
      *err = lerr;
    else
      ILU_HANDLED(lerr);
    ILU_ERR_SWITCH(*err) {
      ILU_SUCCESS_CASE
	0 /* no effect */ ;
      ILU_ERR_CASE(comm_failure, v)
	_ilu_CloseIoingConnection(conn, TRUE);
      ILU_ERR_ELSE
	0 /* no effect */ ;
    } ILU_ERR_ENDSWITCH;
  }
  if (proto != NIL && proto->pr_prefinish_call != NULLFN)
    (void) (*proto->pr_prefinish_call) (call, &lerr);
  if (!_ilu_ReleaseConnIO(conn, FALSE, &lerr))
    goto dun2;
dun4:
  (void) _ilu_ReleaseConnCall(conn, call, FALSE, &lerr);
dun3:
  if (call->ca_msInput && !call->ca_dontWait && !call->ca_disownWait)
    (void) _ilu_ReleaseConnWait(conn, call, FALSE, &lerr);
dun2:
  si = call->ca_si;
  if (si) {
    if (ilu_Check(si->si_busy, err))
      si->si_busy = FALSE;
  }
  _ilu_MaybeFreeSerializer(si);
  if (call->ca_pl) {
    call->ca_pl->pl_nCalls -= 1;
    _ilu_MaybeFreePipeline(call->ca_pl);
    call->ca_pl = NIL;
  }
  if (conn) {
    conn->co_nCalls -= 1;
    if (conn->co_nCalls == 0)
      conn->co_pipeline = NIL;
    if (conn->co_doomed && conn->co_nCalls == 0 && ILU_ERROK(lerr)) {
      lerr = _ilu_CloseConnection(conn);
    }
    rewatch = (!connection_closed(conn) && !connection_incoming(conn)
	       && !_ilu_CanCondition() && conn->co_nOuts == 0);
    _ilu_MaybeFreeConnection(conn);
  }
  (void) _ilu_ExitServerMutex(s, TRUE, &lerr);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, &lerr);
dun0:
  if (ILU_ERROK(*err))
    *err = lerr;
  else
    ILU_HANDLED(lerr);
  if (proto != NIL && proto->pr_finish_call != NULLFN)
    (void) (*proto->pr_finish_call) (call, &lerr);
  if (ILU_ERROK(*err))
    *err = lerr;
  else
    ILU_HANDLED(lerr);
  if (rewatch) {
    (void) ilu_SetConnectionInputHandler(conn, ReadExtraMsg, conn, &lerr);
    if (ILU_ERROK(*err))
      *err = lerr;
    else
      ILU_HANDLED(lerr);
  }
  if (call->ca_incoming && (call->ca_caller != NIL)) {
    ilu_DestroyPassport(call->ca_caller, &lerr);
    call->ca_caller = NIL;
    if (ILU_ERROK(*err))
      *err = lerr;
    else
      ILU_HANDLED(lerr);
  }
  return;
}

/*L1, L2 unconstrained*/
static ilu_RcvReqStat Unbuild(ilu_Call call, ilu_RcvReqStat ans)
{
  return ans;
}

/*Main Invariant holds*/
/**before: L2 disjoint {conn's callmu, iomu, waitmu},
 *  after: Call-Locking(call, IHi)                   if  *initted,
 *  after: L2 disjoint {conn's callmu, iomu, waitmu} if !*initted */

ilu_RcvReqStat
ilu_ReceiveRequest(ilu_Call_s * call, ilu_boolean * initted,
		   ilu_Connection conn, ilu_Class * pit,
		   ilu_Method * meth, ilu_cardinal * sn,
		   ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Server      server = connection_server(conn);
  ilu_PacketType  pkt_type;
  ilu_Error       lerr = ILU_INIT_NO_ERR;
  ilu_Protocol    proto = conn->co_protocol;
  ilu_ReadHeaderResultCode rhrc = ilu_rhrc_error;
  ilu_boolean     stdMeth = FALSE;

  BuildCall(call, conn, server, 0, TRUE);
  call->ca_disownWait = FALSE;
  *initted = FALSE;
  if (!ilu_EnterMutex(ilu_cmu, err))
    return Unbuild(call, ilu_RcvReqStat_noop);
  if (!_ilu_EnterServerMutex(server, FALSE, err)) {
    (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
    return Unbuild(call, ilu_RcvReqStat_noop);
  }
  if (connection_closed(conn)) {
    (void) _ilu_ExitServerMutex(server, TRUE, err);
    (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
    return Unbuild(call, ilu_RcvReqStat_quit);
  }
  if (conn->co_waiting) {
    ILU_ERR_CONS0(bad_locks, err, (void) 6);
    (void) _ilu_ExitServerMutex(server, TRUE, err);
    (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
    return Unbuild(call, ilu_RcvReqStat_noop);
  }
  _ilu_TakeConnWait(conn, call);
  if (!_ilu_EnterConnCall(conn, call, FALSE, err))
    goto lab1;
  if (!_ilu_EnterConnIO(conn, FALSE, err))
    goto lab2;
  if (connection_closed(conn)) {
    (void) _ilu_ReleaseConnIO(conn, TRUE, err);
    (void) _ilu_ReleaseConnCall(conn, call, TRUE, err);
    (void) _ilu_ReleaseConnWait(conn, call, TRUE, err);
    (void) _ilu_ExitServerMutex(server, TRUE, err);
    (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
    return Unbuild(call, ilu_RcvReqStat_quit);
  }
  call->ca_ms = ilu_cmsHi;
  call->ca_msInput = TRUE;
  call->ca_dontWait = FALSE;
  if ((*proto->pr_init_call) (call, err)) {
    /* We're now obliged to eventually call pr_finish_call */
    *initted = TRUE;
    conn->co_nCalls += 1;
    goto lab0;
  }
  (void) _ilu_ReleaseConnIO(conn, TRUE, err);
lab2:
  call->ca_ms = ilu_cmsNo;
  call->ca_msInput = FALSE;
  (void) _ilu_ReleaseConnCall(conn, call, TRUE, err);
lab1:
  (void) _ilu_ReleaseConnWait(conn, call, TRUE, err);
lab0:
  (void) _ilu_ExitServerMutex(server, TRUE, err);
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
  ILU_ERR_SWITCH(*err) {
    ILU_SUCCESS_CASE
      {};
    ILU_ERR_CASE3(comm_failure, bad_param, no_memory) {
      return ilu_RcvReqStat_quit;
    }
    ILU_ERR_ELSE
      return ilu_RcvReqStat_noop;
  } ILU_ERR_ENDSWITCH;    
  /* Call-Locking(call, IHi) */
  rhrc = protocol_read_header(proto, call, &pkt_type, sn, &lerr);
  switch (rhrc) {
  case ilu_rhrc_ok:
    call->ca_ios = ilu_ciosIn;
    break;
  case ilu_rhrc_eof:
    (void) _ilu_CloseConnWithIo(conn, FALSE, err);
    return ilu_RcvReqStat_quit;
  case ilu_rhrc_nothing:
  case ilu_rhrc_handled:
    return ilu_RcvReqStat_noop;
  case ilu_rhrc_error:
    ILU_ERR_SWITCH(lerr) {
      ILU_ERR_CASE(comm_failure, v) {
	*err = lerr;
	_ilu_CloseIoingConnection(conn, TRUE);
	return ilu_RcvReqStat_quit;
      }
      ILU_ERR_CASE(bad_param, v) {
	ILU_HANDLED(lerr);
	ILU_ERR_CONS1(internal, err, minor, ilu_im_unhandled, 0);
	return ilu_RcvReqStat_noop;
      }
      ILU_ERR_CASE3(no_memory, internal, broken_locks) {
	*err = lerr;
	return ilu_RcvReqStat_noop;
      }
      ILU_ERR_CASE2(imp_limit, marshal) {
	*err = lerr;
	return ilu_RcvReqStat_noop;
      }
    } ILU_ERR_ENDSWITCH;
  default:
    if (!ilu_Check(FALSE, err))
      return ilu_RcvReqStat_noop;
  }
  if (pkt_type != ilu_PacketType_Request)
    return ilu_RcvReqStat_noop;
  call_serial_number(call) = *sn;
  if (_ilu_AddConnIdentities (call, err), ILU_ERRNOK(*err))
    {
      call->ca_caller = NIL;
      return ilu_RcvReqStat_noop;
    }
  if (protocol_interpret_request(proto, call, err)) {
    ilu_Method      m = call_method(call);
    stdMeth = m != NIL && method_id(m) >= 0xFF00 && method_id(m) < 0xFFFF;
    ILU_NOTE((CALL_DEBUG | INCOMING_DEBUG),
	  ("%-20.20s(%p over %p from %s #%lu, %s.%s)\n",
	   "ilu_ReceiveRequest", call,
	   conn, call_connection_id(call),
	   (unsigned long) call_serial_number(call),
	   class_name(call_intro_type(call)),
	   method_name(call_method(call))));
  } else {
    ILU_NOTE(INCOMING_DEBUG,
	  ("%-20.20s(%p over %p from %s #%lu) error:  *err=%s, ca_pe=%s.\n",
	   "ilu_ReceiveRequest", call, conn, call_connection_id(call),
	   (long unsigned) call_serial_number(call),
	   ILU_ERR_NAME(*err), ilu_PEName(call->ca_pe)));
    (void) ilu_Check((ILU_ERRNOK(*err) ||
		      call->ca_pe != ilu_ProtocolException_Success),
		     err);
    return ilu_RcvReqStat_noop;
  }
  if (!_ilu_EnterServerMutex(server, TRUE, err))
    return ilu_RcvReqStat_noop;
  if (conn->co_port->po_call_cache != NIL) {
    ilu_CachedCall *cc = conn->co_port->po_call_cache;
    int             i;
    for (i = 0; i < conn->co_port->po_call_cache_size; i++) {
      if ((cc[i].cc_sn == *sn)
	  && (cc[i].cc_intro_type == call->ca_intro_type)
	  && (cc[i].cc_meth == call_method(call))
	  && (cc[i].cc_replyMsg.msg_base != NIL)
	  && strcmp(cc[i].cc_peerinfo, conn_peerinfo(conn)) == 0) {
	if (!_ilu_ExitServerMutex(server, TRUE, err))
	  return ilu_RcvReqStat_noop;
	ILU_NOTE(CALL_DEBUG,
	      ("ilu_ReceiveRequest(%p): resending cached reply"
	       " to call %ld from %s.\n",
	       call, cc[i].cc_sn, cc[i].cc_peerinfo));
	call->ca_ios = ilu_ciosNone;
	if (!protocol_discard_input(proto, call, err)) {
	  return ilu_RcvReqStat_noop;
	}
	if (!transport_send_whole_message(conn->co_transport,
					  &cc[i].cc_replyMsg,
					  err))
	  return ilu_RcvReqStat_noop;
	return ilu_RcvReqStat_noop;
      }
    }
  }
  if (!_ilu_ExitServerMutex(server, TRUE, err))
    return ilu_RcvReqStat_noop;
  if (stdMeth) {
    (call_method(call)->me_stubprocs[_ilu_InternalLanguageIndex]) (call);
    return ilu_RcvReqStat_noop;
  }
  *pit = call->ca_intro_type;
  *meth = call_method(call);
  call->ca_pe = ilu_ProtocolException_GarbageArguments;
  return ilu_RcvReqStat_request;
}

#ifdef WIN32
/* AVOID COMPILER BUG by turning off global optimization for this function
  F:\ilu\src\runtime\kernel\call.c(457) : fatal error C1001: INTERNAL COMPILER ERROR
  (compiler file 'l:\b_bld\c2\P2\main.c', line 374) */
#pragma optimize("g", off)
#endif

/*L1, L2, Main unconstrained*/
static ilu_FineTime ClipAddTime(ilu_FineTime a, ilu_FineTime b,
				ilu_FineTime l)
{
  ilu_FineTime c = ilu_FineTime_Add(a, b);
  if (ilu_FineTime_Cmp(c, l) > 0)
       return l;
  else return c;
}

#ifdef WIN32
/* restore optimizations to original settings */
#pragma optimize("", on)
#endif

/*mxamu = ilu_cmu*/
typedef struct {
  /*L1 >= {ilu_cmu}; L2 unconstrained*/
  
  ilu_Alarmette_s	gra_alarmette;
  ilu_private		gra_cc;
} GetReplyAlarm;

/*L1_sup = ilu_cmu; L2 unconstrained*/

static void GRInvoke(ilu_Alarmette a);
static void GRSet(ilu_FineTime t);
static void GRCancel(void);

static ilu_Alarmette_s grHead = {&grHead, &grHead, FALSE, {0, 0}};
static ilu_AlarmRep grar = {&grHead, GRInvoke, GRSet, GRCancel};

/*Main Invariant holds*/
static void GraInvoke(ilu_private rock)
{
  ilu_FineTime now = ilu_FineTime_Now();
  _ilu_AcquireMutex(ilu_cmu);
  ilu_MXAProc(now, &grar);
  _ilu_ReleaseMutex(ilu_cmu);
  return;
}

static void GRInvoke(ilu_Alarmette a)
{
  ilu_Error err;
  GetReplyAlarm *gra = (GetReplyAlarm*) a;
  err = _ilu_NotifyCondition(gra->gra_cc);
  ILU_MUST_BE_SUCCESS(err);
  return;
}

static void GRSet(ilu_FineTime t)
{
  ilu_SetAlarm(_ilu_grAlarm, t, GraInvoke, NIL);
  return;
}

static void GRCancel(void)
{
  ilu_UnsetAlarm(_ilu_grAlarm);
  return;
}

/*L1, L2 unconstrained*/
static void FreeMessage(ilu_Message * msg)
{
  if (msg->msg_base != NIL) {
    ilu_free(msg->msg_base);
    msg->msg_base = NIL;
  }
  return;
}

/*L1 >= {server}; L2 unconstrained */
static ilu_ReplyList 
GetQueued(ilu_Call call)
{
  ilu_Connection  conn = call_connection(call);
  ilu_ReplyList   r, *pr;
  for (pr = &conn->co_replies, r = *pr;
       r != NIL;
       pr = &r->rp_next, r = *pr)
    if (r->rp_SN == call_serial_number(call)) {
      *pr = r->rp_next;
      return r;
    }
  return NIL;
}

/**Before: Call-Locking(call, OHi);
    After: Call-invariant(call, err) &&
	   (success => Call-Locking(call, IHi)) */
ilu_ProtocolException
ilu_GetReply(ilu_Call call, ilu_cardinal * estatus,
	     ilu_Connection * new_conn,
	     ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Connection  conn = call_connection(call);
  ilu_Protocol    proto = connection_protocol(conn);
  ilu_Server      server = connection_server(conn);
  ilu_Transport   trans = connection_transport(conn);
  ilu_cardinal    pktsn;
  ilu_PacketType  pkttype;
  ilu_ProtocolException ans;
  ilu_boolean     reliable = transport_reliable(trans);
  ilu_boolean     prco = protocol_concurrent(proto);
  ilu_boolean     MT = _ilu_CanCondition();
  ilu_FineTime    now;
  ilu_FineTime    timeout;
  ilu_FineTime    limit, *elimit;
  ilu_FineTime    grandLimit;
  GetReplyAlarm   gra = {{NIL, NIL, FALSE, {0, 0}}, NIL};
  const char     *reply_src = NIL;
  ilu_ProtocolInfo new_pinfo = NIL;
  ilu_Protocol    new_proto = NIL;
  ilu_TransportInfo new_tinfo = NIL;
  ilu_TransportCreator new_tcr = NIL;
  ilu_RelocateScope rel_scope;
  ilu_boolean     rewatch = FALSE;
  ilu_Connection  oldConn = conn;

  *new_conn = NIL;
  if (method_asynchronous(call_method(call)))
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh,
			 ilu_ProtocolException_Not);
  if (call->ca_dontWait)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh,
			 ilu_ProtocolException_Not);
  if (conn->co_serialer != call->ca_si)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh,
			 ilu_ProtocolException_Not);
  if (conn->co_mucall != call || !conn->co_ioing || call->ca_msInput)
    return ILU_ERR_CONS0(bad_locks, err, ilu_ProtocolException_Not);

  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p over %p to \"%s\" #%lu)...\n",
			"ilu_GetReply", call, call_connection(call),
			call_server_id(call),
			(unsigned long) call_serial_number(call)));

  timeout = trans->tr_to1;
  if (!reliable) {
    now = ilu_FineTime_Now();
    limit = ilu_FineTime_Add(now, timeout);
    grandLimit = ilu_FineTime_Add(now, trans->tr_tto);
    elimit = &limit;
  } else
    elimit = NIL;
  /* Call-Locking(call, OHi) */
  /* (=> (this thread holds waitmu iff disownWait)) */
  while (1) {
    /* Call-Locking(call), IHi or OHi; !call->ca_dontWait */
    ilu_ReplyList   qrl = NIL;
    if (!ilu_EnterMutex(ilu_cmu, err))
      return ilu_ProtocolException_Not;
    if (!_ilu_EnterServerMutex(server, FALSE, err))
      goto abort1;
    if (connection_closed(conn)) {
      ILU_ERR_CONS1(comm_failure, err, minor, ilu_cfm_conn_lost,
		    (void) 6);
      goto abort2;
    }
    if (!_ilu_ReleaseConnIO(conn, FALSE, err))
      goto abort2;
    if (prco || conn->co_pipeline) {	/* release the call mutex */
      conn->co_nOuts++;
      if (!_ilu_ReleaseConnCall(conn, call, FALSE, err))
	goto abort2;
    }
    call->ca_ms = ilu_cmsLo;
    /* L1 = {cmu, server}; Call-Lo(call) */
    /* L2 >= {conn's waitmu} iff ca_disownWait || ca_msInput */
    if (MT) {
      if (!reliable) {
	gra.gra_cc = connection_cc(conn);
	ilu_MXASet(&grar, &gra.gra_alarmette, limit);
      }
      if (conn->co_waiting && !call->ca_msInput) {
	while (conn->co_waiting) {
	  if (!ilu_CMWait2(connection_cc(conn), server_lock(server),
			   ilu_cmu, err)) {
	    if (!reliable)
	      ilu_MXAClear(&grar, &gra.gra_alarmette);
	    goto abort2;
	  }
	  qrl = GetQueued(call);
	  if (qrl)
	    goto waitdun;
	}
	_ilu_TakeConnWait(conn, call);
	goto waitdun;
      }
    }
    if ((MT && !call->ca_msInput) || !conn->co_waiting)
      _ilu_TakeConnWait(conn, call);
    (void) _ilu_BlockingWaitForInputOnConnection(conn, elimit,
						 err);
    ILU_ERR_SWITCH(*err) {
      ILU_SUCCESS_CASE
	0;
      ILU_ERR_CASE2(interrupted, internal)
	0;
      ILU_ERR_CASE(broken_locks, x)
	return ilu_ProtocolException_Not;
    } ILU_ERR_ENDSWITCH;
waitdun:
    /* L1 = {cmu, server}; Call-Lo(call) */
    /* L2 >= {conn's waitmu} iff !qrl */
    if (MT && !reliable)
      ilu_MXAClear(&grar, &gra.gra_alarmette);
    if (call->ca_irq || ILU_ERRNOK(*err)) {
      if (!call->ca_disownWait && !qrl
	  && !_ilu_ReleaseConnWait(conn, call, TRUE, err))
	return ilu_ProtocolException_Not;
      qrl->rp_next = conn->co_replies;
      /*
       * Somebody else (ilu_FinishCall?) should (but, doesn't yet)
       * take care of reaping these.
       */
      conn->co_replies = qrl;
      if (!_ilu_ExitServerMutex(server, TRUE, err))
	return ilu_ProtocolException_Not;
      if (!ilu_ExitMutex(ilu_cmu, TRUE, err))
	return ilu_ProtocolException_Not;
      if (ILU_ERROK(*err))
	ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
      return ilu_ProtocolException_Not;
    }
    if (prco || conn->co_pipeline) {	/* re-acquire the call mutex */
      _ilu_Assert(conn->co_nOuts > 0, "GetReply: conn->co_nOuts==0 a");
      if (!_ilu_EnterConnCall(conn, call, TRUE, err))
	return ilu_ProtocolException_Not;
      _ilu_Assert(conn->co_nOuts > 0, "GetReply: conn->co_nOuts==0 b");
      conn->co_nOuts--;
    }
    if (!_ilu_TakeConnIO(conn, TRUE, err))	/* re-acquire the I/O
						 * mutex */
      return ilu_ProtocolException_Not;
    call->ca_ms = ilu_cmsHi;
    call->ca_msInput = TRUE;
    call->ca_dontWait = (qrl != NIL);
    /* L1=={cmu, server} && Call-Locking-Remnant(call, IHi) */
    if (connection_closish(conn) && !qrl) {
      if (!_ilu_ExitServerMutex(server, TRUE, err))
	return ilu_ProtocolException_Not;
      if (!ilu_ExitMutex(ilu_cmu, TRUE, err))
	return ilu_ProtocolException_Not;
      return ilu_ProtocolException_LostConnection;
    }
    if ((prco || conn->co_pipeline) && !qrl) {
      qrl = GetQueued(call);
      call->ca_dontWait = (qrl != NIL);
      if (qrl && !call->ca_disownWait) {
	(void) _ilu_ReleaseConnWait(conn, call, TRUE, err);
	ILU_MUST_BE_SUCCESS(*err);
      }
    }
    if (!_ilu_ExitServerMutex(server, TRUE, err))
      return ilu_ProtocolException_Not;
    if (!ilu_ExitMutex(ilu_cmu, TRUE, err))
      return ilu_ProtocolException_Not;
    /* Call-Locking(call, IHi) */
    if (qrl) {
      FreeMessage(&call->ca_msg);
      protocol_resume_interp(proto, call, qrl->rp_queued);
      call->ca_ios = ilu_ciosIn;
      ilu_free(qrl);
      reply_src = "queue";
      goto interpd;
    }
    while (1) {
      /* Read all available replies */
      /* Call-Locking(call, IHi); !call->ca_dontWait */
      ilu_boolean     closeit = FALSE;
      ilu_ReadHeaderResultCode rhrc;
      rhrc = protocol_read_header(proto, call, &pkttype, &pktsn, err);
      switch (rhrc) {
      case ilu_rhrc_ok:
	call->ca_ios = ilu_ciosIn;
	break;
      case ilu_rhrc_nothing:
	goto repliesRead;
      case ilu_rhrc_handled:
	break;
      case ilu_rhrc_eof:
	ILU_NOTE(CONNECTION_DEBUG,
		 ("Closing connection %p to %s due to EOF.\n",
		  conn, server->sr_id));
	closeit = TRUE;
      case ilu_rhrc_error:
	ILU_ERR_SWITCH(*err) {
	  ILU_ERR_CASE(bad_param, v) {
	    ILU_HANDLED(*err);
	    (void) ILU_ERR_CONS1(internal, err, minor, ilu_im_unhandled,
				 0);
	  }
	  ILU_ERR_CASE(comm_failure, v) {
	    ILU_NOTE(CONNECTION_DEBUG,
		   ("%s %p to %s due to comm_failure, minor=%d.\n",
		    "Closing connection", conn,
		    server->sr_id, v->minor));
	    closeit = TRUE;
	  }
	  ILU_ERR_ELSE
	    0 /* statement with no effect */ ;
	} ILU_ERR_ENDSWITCH;
	if (!closeit)
	  return ilu_ProtocolException_Not;
	else
	  break;
      default:
	_ilu_Assert(FALSE, "GetReply vs read_header");
      }
      if (closeit) {
	if (!ilu_ReEnterMutex(ilu_cmu, err))
	  return ilu_ProtocolException_Not;
	if (!_ilu_EnterServerMutex(server, TRUE, err))
	  return ilu_ProtocolException_Not;
	_ilu_CloseIoingConnection(conn, ilu_TRUE);
	if (!_ilu_ExitServerMutex(server, TRUE, err))
	  return ilu_ProtocolException_Not;
	if (!ilu_ExitMutex(ilu_cmu, TRUE, err))
	  return ilu_ProtocolException_Not;
	return ilu_ProtocolException_LostConnection;
      }
      if (pkttype == ilu_PacketType_Reply) {
	if (pktsn == call_serial_number(call)) {
	  FreeMessage(&call->ca_msg);
	  reply_src = "wire";
	  goto interpd;
	}
	ILU_NOTE((CALL_DEBUG | INCOMING_DEBUG),
		 ("%-20.20s(%p over %p to \"%s\" #%lu)"
		  " %s reply for SN #%lu.\n",
		  "ilu_GetReply", call, call_connection(call),
		  call_server_id(call),
		  (unsigned long) call_serial_number(call),
		  (prco || conn->co_pipeline) ? "queueing" : "discarding",
		  (unsigned long) pktsn));
	if ((prco || conn->co_pipeline)) {
	  ilu_ReplyList   r;
	  ilu_refany      queued2;
	  call->ca_ios = ilu_ciosNone;
	  queued2 = protocol_delay_interp(proto, call, err);
	  if (ILU_ERRNOK(*err))
	    return ilu_ProtocolException_Not;
	  r = (ilu_ReplyList) ilu_MallocE(sizeof(*r), err);
	  if (ILU_ERRNOK(*err))
	    return ilu_ProtocolException_Not;
	  r->rp_queued = queued2;
	  r->rp_SN = pktsn;
	  if (!_ilu_EnterServerMutex(server, TRUE, err))
	    return ilu_ProtocolException_Not;
	  r->rp_next = conn->co_replies;
	  conn->co_replies = r;
	  if (!_ilu_ExitServerMutex(server, TRUE, err))
	    return ilu_ProtocolException_Not;
	} else {
	  /* Reply to different call; discard. */
	  call->ca_ios = ilu_ciosNone;
	  if (!protocol_discard_input(proto, call, err))
	    return ilu_ProtocolException_Not;
	}
      } else {
	/* not a reply */
	call->ca_ios = ilu_ciosNone;
	if (!protocol_discard_input(proto, call, err))
	  return ilu_ProtocolException_Not;
      }
    }
repliesRead:
    /* Call-Locking(call, IHi) */
    ILU_MUST_BE_SUCCESS(*err);
    if (!reliable) {
      now = ilu_FineTime_Now();
      if (ilu_FineTime_Cmp(now, grandLimit) >= 0)
	break;
      if (ilu_FineTime_Cmp(now, limit) >= 0) {
	ILU_NOTE(CONNECTION_DEBUG,
	  ("GetReply: retransmitting request on conn %p\n", conn));
	if (!transport_send_whole_message(trans, &call->ca_msg, err))
	  return ilu_ProtocolException_Not;
	timeout = ClipAddTime(timeout, timeout, trans->tr_toN);
	limit = ClipAddTime(now, timeout, grandLimit);
      }
    }
  }
  return ilu_ProtocolException_RequestTimeout;
interpd:
  /* Call-Locking(call, IHi) */
  conn->co_lastWorking = FALSE;
  ans = protocol_interpret_reply(proto, call, estatus, err);
  ILU_ERR_SWITCH(*err) {
    ILU_ERR_CASE(relocate, rel) {
      ILU_ERRS((no_memory, internal/check)) merr;
      ilu_string      ts = _ilu_StringifyTinfo(rel->rel_tinfo, &merr);
      static char    *noT = "(not enough memory to stringify transport info)";
      if (!ILU_ERROK(merr))
	ts = noT;
      ILU_NOTE((CALL_DEBUG | INCOMING_DEBUG),
	 ("%-20.20s(%p over %p to \"%s\" #%lu try %u) => (from %s)"
	  " ans=%s estatus=%lu err=relocate(%s, %s, %s)\n",
	  "ilu_GetReply", call, call_connection(call),
	  call_server_id(call),
	  (unsigned long) call_serial_number(call),
	  call->ca_tryIndex,
	  reply_src, ilu_PEName(ans),
	  (unsigned long) *estatus,
	  ilu_RelocateScope_Name(rel->rel_scope),
	  rel->rel_pinfo, ts));
      if (ts != noT)
	ilu_free(ts);
      ILU_HANDLED(merr);
      if (call->ca_tryIndex < 4) {
	new_pinfo = rel->rel_pinfo;
	new_tinfo = rel->rel_tinfo;
	rel_scope = rel->rel_scope;
	rel->rel_pinfo = NIL;
	rel->rel_tinfo = NIL;
	ILU_HANDLED(*err);
	goto dorelocate;
      } else {
	ILU_HANDLED(*err);
	ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_relocate_loop, 0);
      }
    }
    ILU_ERR_ELSE
      0;
  } ILU_ERR_ENDSWITCH;
  ILU_NOTE((CALL_DEBUG | INCOMING_DEBUG),
	   ("%-20.20s(%p over %p to \"%s\" #%lu) => (from %s)"
	    " ans=%s estatus=%lu err=%s\n",
	    "ilu_GetReply", call, call_connection(call),
	    call_server_id(call),
	    (unsigned long) call_serial_number(call),
	    reply_src, ilu_PEName(ans), (unsigned long) *estatus,
	    ILU_ERR_NAME(*err)));
  return ans;
dorelocate:
  {
    /* Call-Locking(call, IHi) */
    ilu_Connection  cur;
    int             openres;
    call->ca_ios = ilu_ciosNone;
    if (!ilu_ReEnterMutex(ilu_cmu, err))
      return ilu_ProtocolException_Not;
    if (!_ilu_EnterServerMutex(server, TRUE, err))
      return ilu_ProtocolException_Not;
    if (rel_scope != ilu_relocate_call)
      conn->co_doomed = TRUE;
    if (proto != NIL && proto->pr_prefinish_call != NULLFN)
      (void) (*proto->pr_prefinish_call) (call, err);
    _ilu_ReleaseConnIO(conn, TRUE, err);
    _ilu_ReleaseConnCall(conn, call, TRUE, err);
    if (!call->ca_dontWait && !call->ca_disownWait)
      (void) _ilu_ReleaseConnWait(conn, call, TRUE, err);
    call->ca_ms = ilu_cmsNo;
    conn->co_nCalls -= 1;
    if (conn->co_nCalls == 0)
      conn->co_pipeline = NIL;
    if (conn->co_doomed && conn->co_nCalls == 0 && ILU_ERROK(*err)) {
      *err = _ilu_CloseConnection(conn);
    }
    rewatch = (!connection_closed(conn) && !connection_incoming(conn)
	       && !_ilu_CanCondition() && conn->co_nOuts == 0);
    if (conn->co_serialer) {
      ilu_Serializer  si = conn->co_serialer;
      conn->co_serialer = NIL;
      si->si_conn = NIL;
      ILU_NOTE(CONNECTION_DEBUG,
	       ("ilu_GetReply(%p): dissociating serializer %p"
		" and connection %p.\n",
		call, si, conn));
    }
    _ilu_MaybeFreeConnection(conn);
    conn = call->ca_connection = NIL;
    call->ca_disownWait = FALSE;
    if (ILU_ERRNOK(*err))
      goto abort2;
    if (rewatch) {
      (void) _ilu_InnerSetConnectionInputHandler(oldConn, ReadExtraMsg,
						 oldConn, err);
      if (ILU_ERRNOK(*err))
	goto abort2;
    }
    for (cur = server_connections(server);
	 cur != NIL && !conn;
	 cur = connection_next(cur)) {
      if (!connection_incoming(cur) &&
	  strcmp(new_pinfo, cur->co_pinfo) == 0 &&
	  _ilu_CompareTinfo(new_tinfo, conn_tinfo(cur))) {
	if (!cur->co_mucall && !cur->co_serialer &&
	    !cur->co_doomed && !cur->co_closing &&
	    (!call->ca_si || !connection_concurrent(cur)) &&
	    (cur->co_pipeline == NIL
	     || cur->co_pipeline == call->ca_pl))
	  conn = cur;
      }
    }
    if (conn)
      openres = 3;
    else {
      new_tcr = _ilu_GetTransportCreator(new_tinfo, err);
      if (ILU_ERRNOK(*err))
	goto abort2;
      new_proto = _ilu_GetProtocolFromInfo(new_pinfo);
      if (!proto) {
	ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_ps, (void) 6);
	goto abort2;
      }
      openres = OpenConn(new_pinfo, new_proto, new_tinfo, new_tcr,
			 server, call->ca_caller, &conn, new_conn,
			 err);
      if (conn && rel_scope == ilu_relocate_call)
	conn->co_doomed = TRUE;
      _ilu_CloseTCR(new_tcr);
    }
    switch (openres) {
    case 3:
      call->ca_connection = conn;
      (void) _ilu_EnterConnCall(conn, call, TRUE, err);
      (void) _ilu_TakeConnIO(conn, TRUE, err);
      if (ILU_ERROK(*err)) {
	call->ca_ms = ilu_cmsHi;
	call->ca_msInput = FALSE;
	conn->co_nCalls += 1;
	if (call->ca_si) {
	  conn->co_serialer = call->ca_si;
	  call->ca_si->si_conn = conn;
	  ILU_NOTE(CONNECTION_DEBUG,
		   ("ilu_GetReply(%p): associating serializer %p"
		    " and connection %p.\n",
		    call, call->ca_si, conn));
	}
	conn->co_pipeline == call->ca_pl;
	call->ca_tryIndex += 1;
	if ((*proto->pr_init_call) (call, err))
	  ILU_ERR_CONS1(transient, err, minor, ilu_tm_retry, (void) 0);
      }
    case 2:
      (void) _ilu_ExitServerMutex(server, TRUE, err);
    case 1:
      (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
    case 0:
      return ilu_ProtocolException_Not;
    }
  }
abort2:
  (void) _ilu_ExitServerMutex(server, TRUE, err);
abort1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
  return ilu_ProtocolException_Not;
}


/*Main Invariant holds; L2 disjoint {conn's callmu, iomu, waitmu}*/
ilu_boolean
ilu_OutgoingConnectionThreadProc(ilu_Connection conn,
				 ILU_ERRS((IoErrs)) * err)
{
  ilu_Protocol    proto = connection_protocol(conn);
  ilu_Server      server = connection_server(conn);
  ilu_boolean     prco = protocol_concurrent(proto);
  ilu_Call_s      dummyCall;
  ILU_ERRS((interrupted)) lerr;
  ILU_ERRS((IoErrs)) ferr;
  ilu_boolean     ok;
  if (!ilu_CanCondition() || connection_incoming(conn) ||
      !conn->co_lsrCares)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  BuildCall(&dummyCall, conn, server, 0, FALSE);
  dummyCall.ca_disownWait = FALSE;
  if (!(*proto->pr_init_call) (&dummyCall, err))
    return FALSE;
  if (!ilu_EnterMutex(ilu_cmu, err))
    goto dun0;
  if (!_ilu_EnterServerMutex(server, FALSE, err))
    goto dun1;
  while (1) {
    /* L1 = {cmu, server}; Main Remnant; L2 disjoint conn's mutexes */
    while (1) {			/* wait for input outside of a call */
      while ((conn->co_nOuts || conn->co_waiting)
	     && !connection_closed(conn))
	if (!ilu_CMWait2(connection_cc(conn), server_lock(server),
			 ilu_cmu, &lerr))
	  goto dun2;
      if (connection_closed(conn))
	goto dun2;
      _ilu_TakeConnWait(conn, &dummyCall);
      ok = _ilu_BlockingWaitForInputOnConnection(conn, NIL, &lerr);
      if (!ilu_Check(ILU_ERROK(lerr), err))
	/* LSR has no reason to abort this thread */
	goto dun3;
      if (!prco)
	if (!_ilu_ReleaseConnWait(conn, &dummyCall, TRUE, err))
	  goto dun2;
      if (!_ilu_EnterConnCall(conn, &dummyCall, TRUE, err)) {
	if (prco)
	  goto dun3;
	else
	  goto dun2;
      }
      if (!prco)
	_ilu_TakeConnWait(conn, &dummyCall);
      if (!_ilu_EnterConnIO(conn, TRUE, err))
	goto dun4;
      if (connection_closed(conn))
	goto dun5;
      if (conn->co_nOuts == 0)
	goto doit;
      if (!_ilu_ReleaseConnIO(conn, TRUE, err))
	goto dun4;
      if (!_ilu_ReleaseConnCall(conn, &dummyCall, TRUE, err))
	goto dun3;
      if (!_ilu_ReleaseConnWait(conn, &dummyCall, TRUE, err))
	goto dun2;
    }
doit:
    /* L1 = {cmu, server}; L2 >= {conn's callmu, iomu, waitmu} */
    if (!ProcessExtraInput(conn, &dummyCall, err))
      goto dun5;
    if (!_ilu_ReleaseConnIO(conn, TRUE, err))
      goto dun4;
    if (!_ilu_ReleaseConnCall(conn, &dummyCall, TRUE, err))
      goto dun3;
    if (!_ilu_ReleaseConnWait(conn, &dummyCall, TRUE, err))
      goto dun2;
  }
dun5:
  if (ILU_ERROK(*err) && proto->pr_prefinish_call != NULLFN)
    (void) (*proto->pr_prefinish_call) (&dummyCall, err);
#ifdef ENABLE_DEBUGGING
  {
    ilu_string      t;
    static ilu_string fallback = "(not enough memory to produce)";
    ILU_ERRS((no_memory, internal/check)) serr;
    t = _ilu_StringifyTinfo(conn_tinfo(conn), &serr);
    if (!t)
      t = fallback;
    ILU_NOTE(CONNECTION_DEBUG,
	  ("%s(%p, tr=%p, %s over %s): done, closing.\n",
	   "ilu_OutgoingConnectionThreadProc", conn,
	   conn->co_transport, conn->co_pinfo, t));
    if (t != fallback)
      ilu_free(t);
    if (ILU_ERROK(*err))
      *err = serr;
    else
      ILU_HANDLED(serr);
  }
#endif
  _ilu_CloseIoingConnection(conn, FALSE);
  (void) _ilu_ReleaseConnIO(conn, TRUE, err);
dun4:
  (void) _ilu_ReleaseConnCall(conn, &dummyCall, TRUE, err);
dun3:
  (void) _ilu_ReleaseConnWait(conn, &dummyCall, TRUE, err);
dun2:
  conn->co_lsrCares = FALSE;
  _ilu_MaybeFreeConnection(conn);
  (void) _ilu_ExitServerMutex(server, TRUE, err);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
dun0:
  (*proto->pr_finish_call) (&dummyCall, &ferr);
  if (ILU_ERROK(*err))
    *err = ferr;
  else
    ILU_HANDLED(ferr);
  return ILU_ERROK(*err);
}

/* Main holds; L2 not further constrained */
ilu_boolean
ilu_InterruptCall(ilu_Call call,
		  ILU_ERRS((bad_locks, broken_locks,
			    bad_param)) * err)
{
  ilu_Connection  conn = call ? call_connection(call) : NIL;
  ilu_Server      s = call ? call_server(call) : NIL;
  ilu_boolean     ans;
  if (_ilu_CanCondition())
    return ILU_ERR_CONS1(bad_param, err, minor,
			 ilu_bpm_threading, FALSE);
  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  call->ca_irq = TRUE;
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p over %p to \"%s\" #%lu)\n",
		     "ilu_InterruptCall", call,
		     call_connection(call), call_server_id(call),
		     (unsigned long) call_serial_number(call)));
  if (!ilu_EnterMutex(ilu_cmu, err))
    return FALSE;
  if (!_ilu_EnterServerMutex(s, FALSE, err))
    return FALSE;
  if (conn->co_closed)
    ans = TRUE;
  else {
    int             forMe = !conn->co_ioing;
    if (forMe)
      if (!_ilu_TakeConnIO(conn, TRUE, err))
	goto dun2;
    /* XXX make [un]marshalling kernel entry pts check ca_irq */
    ans = transport_interruptST(conn->co_transport, err);
    if (forMe)
      if (!_ilu_ReleaseConnIO(conn, TRUE, err))
	goto dun2;
  }
dun2:
  (void) _ilu_ExitServerMutex(s, TRUE, err);
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
  return ans && ILU_ERROK(*err);
}

/*Call-Locking(call, OHi)*/

ilu_boolean
ilu_StartRequest(ilu_Call call, ilu_cardinal argSize,
		 ILU_ERRS((bad_param, IoErrs)) * err)
{
  ilu_Connection  conn = call_connection(call);
  ilu_boolean ans;

  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  ans = protocol_start_request(call_proto(call), call, argSize, err);
  ILU_NOTE(CALL_DEBUG,
	("%-20.20s(%p over %p to \"%s\" #%lu, argSize %lu) => %s\n",
	 "ilu_StartRequest", call, call_connection(call),
	 call_server_id(call),
	 (unsigned long) call_serial_number(call),
	 argSize, ILU_ERR_NAME(*err)));
  if (!ans)
    return FALSE;
  call->ca_ios = ilu_ciosOut;
  return TRUE;
}

ilu_boolean
ilu_FinishRequest(ilu_Call call,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_Connection  conn = call_connection(call);
  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  call->ca_ios = ilu_ciosNone;
  conn->co_lastWorking = TRUE;
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p over %p to \"%s\" #%lu)\n",
		     "ilu_FinishRequest", call, call_connection(call),
		     call_server_id(call),
		     (unsigned long) call_serial_number(call)));
  return protocol_finish_request(call_proto(call), call,
				 &call->ca_msg, err);
}

/**before: Call-Locking(call, IHi);
    after: Call-Invariant(call, err),
	   success => Call-Locking(call, VLo)*/
ilu_boolean
ilu_RequestRead(ilu_Call call,
		ILU_ERRS((IoErrs)) * err)
{
  ilu_Connection  conn = call_connection(call);
  ilu_Server      s = connection_server(conn);
  ilu_Protocol    p = call_proto(call);
  call->ca_ios = ilu_ciosNone;
  call->ca_pe = ilu_ProtocolException_Success;
  protocol_request_read(p, call, err);
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p from %s #%lu) => %s\n",
		     "ilu_RequestRead", call, call_connection_id(call),
		     (unsigned long) call_serial_number(call),
		     ILU_ERR_NAME(*err)));
  if (ILU_ERRNOK(*err))
    return ilu_FALSE;
  if (!ilu_EnterMutex(ilu_cmu, err))
    goto dun0;
  if (!_ilu_EnterServerMutex(s, FALSE, err))
    goto dun1;
  if (!_ilu_ReleaseConnIO(conn, FALSE, err))
    goto dun2;
  if (connection_concurrent(conn)) {
    conn->co_nOuts++;
    (void) _ilu_ReleaseConnCall(conn, call, FALSE, err);
  }
  (void) _ilu_ReleaseConnWait(conn, call, FALSE, err);
  call->ca_ms = ilu_cmsLo;
dun2:
  (void) _ilu_ExitServerMutex(s, TRUE, err);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
dun0:
  return ILU_ERROK(*err);
}

/*Call-Locking(call, IHi)*/

ilu_boolean
ilu_ReplyRead(ilu_Call call,
	      ILU_ERRS((IoErrs)) * err)
{
  ilu_Protocol    p = call_proto(call);
  if (method_asynchronous(call_method(call)))
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  call->ca_ios = ilu_ciosNone;
  protocol_reply_read(p, call, err);
  ILU_NOTE(CALL_DEBUG,
	("%-20.20s(%p over %p to \"%s\" #%lu) => %s\n",
	 "ilu_ReplyRead", call, call_connection(call),
	 call_server_id(call),
	 (unsigned long) call_serial_number(call),
	 ILU_ERR_NAME(*err)));
  return ILU_ERROK(*err);
}

/*Main Invariant holds, L2 otherwise unconstrained*/

ilu_cardinal
ilu_BeginSizingReply(ilu_Call call,
		     ilu_boolean exns_possible,
		     ILU_ERRS((IoErrs)) * err)
{
  ilu_Protocol    proto = call_proto(call);
  if (!protocol_needs_sizing(connection_protocol(call_connection(call))))
    return (ILU_CLER(*err), 0);
  return ((*proto->pr_begin_sizing_reply) (call, exns_possible, err));
}

ilu_cardinal
ilu_BeginSizingException(ilu_Call call,
			 ilu_integer eindex,
			 ILU_ERRS((IoErrs)) * err)
{
  ilu_Protocol    proto = call_proto(call);
  _ilu_Assert(eindex != 0,
	      "BeginSizingException called with zero exceptionVal");
  if (!protocol_needs_sizing(proto))
    return (ILU_CLER(*err), 0);
  return ((*proto->pr_begin_sizing_exn)
	  (call,
	   (eindex > 0) ? eindex : 0,
	   (eindex > 0) ? ilu_ProtocolException_Success
	   : (ilu_ProtocolException) - eindex,
	   err));
}

/**before: Call-Locking(call, Lo);
    after: Call-Invariant(call, err),
	   success => Call-Locking(call, OHi). */

ilu_boolean
ilu_BeginReply(ilu_Call call, ilu_boolean exceptions_p,
	       ilu_cardinal argSize,
	       ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Server      s = call_server(call);
  ilu_Connection  conn = call_connection(call);
  if (!ilu_EnterMutex(ilu_cmu, err))
    goto dun0;
  if (!_ilu_EnterServerMutex(s, FALSE, err))
    goto dun1;
  if (connection_concurrent(conn)
      && !_ilu_EnterConnCall(conn, call, FALSE, err))
    goto dun2;
  if (!_ilu_EnterConnIO(conn, FALSE, err))
    goto dun3;
  if (connection_closed(conn)) {
    (void) ILU_ERR_CONS1(comm_failure, err, minor, ilu_cfm_conn_lost, 6);
    goto dun4;
  }
  if (connection_concurrent(conn)) {
    if (!ilu_Check(conn->co_nOuts > 0, err))
      goto dun2;
    conn->co_nOuts--;
  }
  call->ca_ms = ilu_cmsHi;
  call->ca_msInput = FALSE;
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p from %s #%lu, argSize %lu)\n",
		     "ilu_BeginReply", call, call_connection_id(call),
		     (unsigned long) call_serial_number(call),
		     (unsigned long) argSize));
  if (protocol_begin_reply(call_proto(call), call, exceptions_p,
			   argSize, err))
    call->ca_ios = ilu_ciosOut;
  goto dun2;
dun4:
  (void) _ilu_ReleaseConnIO(conn, TRUE, err);
dun3:
  (void) _ilu_ReleaseConnCall(conn, call, TRUE, err);
dun2:
  (void) _ilu_ExitServerMutex(s, TRUE, err);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
dun0:
  return ILU_ERROK(*err);
}

ilu_boolean
ilu_BeginException(ilu_Call call, ilu_integer exceptionVal,
		   ilu_cardinal argSize,
		   ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Server      s = call_server(call);
  ilu_Connection  conn = call_connection(call);
  ilu_cardinal	exceptionCode = (exceptionVal <= 0) ? 0 : exceptionVal;
  ilu_ProtocolException sys_excn = ((exceptionVal > 0) ?
				    ilu_ProtocolException_Success :
			   (ilu_ProtocolException) - exceptionVal);
  _ilu_Assert(exceptionVal != 0,
	      "BeginException called with zero exceptionVal");
  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  if (!ilu_EnterMutex(ilu_cmu, err))
    goto dun0;
  if (!_ilu_EnterServerMutex(s, FALSE, err))
    goto dun1;
  if (connection_concurrent(conn)
      && !_ilu_EnterConnCall(conn, call, FALSE, err))
    goto dun2;
  if (!_ilu_EnterConnIO(conn, FALSE, err))
    goto dun3;
  if (connection_closed(conn)) {
    (void) ILU_ERR_CONS1(comm_failure, err, minor, ilu_cfm_conn_lost, 6);
    goto dun4;
  }
  if (connection_concurrent(conn)) {
    if (!ilu_Check(conn->co_nOuts > 0, err))
      goto dun2;
    conn->co_nOuts--;
  }
  call->ca_ms = ilu_cmsHi;
  call->ca_msInput = FALSE;
  ILU_NOTE(CALL_DEBUG,
	("%-20.20s(%p from %s #%lu, excn %ld, argSize %lu)\n",
	 "ilu_BeginException", call, call_connection_id(call),
	 (unsigned long) call_serial_number(call),
	 (long) exceptionVal, (unsigned long) argSize));
  if (protocol_begin_exception(call_proto(call), call, exceptionCode,
			       sys_excn, argSize, err))
    call->ca_ios = ilu_ciosOut;
  goto dun2;
dun4:
  (void) _ilu_ReleaseConnIO(conn, TRUE, err);
dun3:
  (void) _ilu_ReleaseConnCall(conn, call, TRUE, err);
dun2:
  (void) _ilu_ExitServerMutex(s, TRUE, err);
dun1:
  (void) ilu_ExitMutex(ilu_cmu, TRUE, err);
dun0:
  return ILU_ERROK(*err);
}

/* Call-Locking(call, OHi) */

ilu_boolean
ilu_FinishReply(ilu_Call call,
		ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Connection  conn = call_connection(call);
  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  call->ca_ios = ilu_ciosNone;
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p from %s #%lu)\n",
		     "ilu_FinishReply", call,
		     call_connection_id(call),
		     (unsigned long) call_serial_number(call)));
  return protocol_finish_reply(call_proto(call), call, err);
}

ilu_boolean 
ilu_FinishException(ilu_Call call,
		    ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Connection  conn = call_connection(call);
  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  call->ca_ios = ilu_ciosNone;
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p from %s #%lu)\n",
		     "ilu_FinishException", call,
		     call_connection_id(call),
		     (unsigned long) call_serial_number(call)));
  return protocol_finish_exception(call_proto(call), call, err);
}

/*L1 >= {call's server}; L2 unconstrained */
ilu_boolean
_ilu_CacheCall(ilu_Call call, ilu_Message * reply,
	       ILU_ERRS((internal)) * err)
{
  ilu_Connection  conn = call_connection(call);
  ilu_Port        p = connection_port(conn);
  ilu_string peerinfo = conn_peerinfo(conn);
  ilu_integer     peerinfo_len = strlen(peerinfo);
  ilu_CachedCall *cc = p->po_call_cache;
  if (peerinfo_len >= MAX_CCPEER_LEN)
    return ILU_ERR_CONS1(internal, err, minor, ilu_im_tInfoLen, FALSE);
  if (cc[p->po_call_cache_finger].cc_replyMsg.msg_base != NIL)
    ilu_free(cc[p->po_call_cache_finger].cc_replyMsg.msg_base);
  strcpy(cc[p->po_call_cache_finger].cc_peerinfo, peerinfo);
  cc[p->po_call_cache_finger].cc_sn = call_serial_number(call);
  cc[p->po_call_cache_finger].cc_intro_type = call_intro_type(call);
  cc[p->po_call_cache_finger].cc_meth = call_method(call);
  cc[p->po_call_cache_finger].cc_replyMsg = *reply;
  p->po_call_cache_finger++;
  p->po_call_cache_finger %= p->po_call_cache_size;
  ILU_CLER(*err);
  return TRUE;
}

/* Call-Locking(call, Lo) */
ilu_boolean 
ilu_NoReply(ilu_Call call,
	    ILU_ERRS((bad_param, bad_locks, broken_locks)) * err)
{
  ilu_Server      s = call_server(call);
  ILU_NOTE(CALL_DEBUG, ("%-20.20s(%p from %s #%lu)\n", "ilu_NoReply",
		     call, call_connection_id(call),
		     (unsigned long) call_serial_number(call)));
  return ILU_CLER(*err);
}

/**For sizing: L1, L2 unconstrained;
   For output: Call-Locking(call, OHi);
   For input:  Call-Locking(call, IHi)*/

/* ==================== optional ==================== */

void ilu_OutputOptional (ilu_Call call, ilu_boolean i,
			 ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_optional(call_proto(call), call, i, the_type, err);
  return;
}

void ilu_InputOptional (ilu_Call call, ilu_boolean *i,
			ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_optional(call_proto(call), call, i, the_type, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfOptional (ilu_Call call, ilu_boolean i,
				 ilu_Type the_type,
				 ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_optional(call_proto(call), call, i, the_type, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== integer ==================== */

void 
ilu_OutputInteger(ilu_Call call, ilu_integer i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_integer(call_proto(call), call, i, err);
  return;
}

void 
ilu_InputInteger(ilu_Call call, ilu_integer * i,
		 ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_integer(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfInteger (ilu_Call call, ilu_integer i,
				      ILU_ERRS((IoErrs)) *err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_integer(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== cardinal ==================== */

void ilu_OutputCardinal (ilu_Call call, ilu_cardinal i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_cardinal(call_proto(call), call, i, err);
  return;
}

void ilu_InputCardinal (ilu_Call call, ilu_cardinal *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_cardinal(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfCardinal (ilu_Call call, ilu_cardinal i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_cardinal(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== character ==================== */

void 
ilu_OutputCharacter(ilu_Call call, ilu_character i,
		    ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_character(call_proto(call), call, i, err);
  return;
}

void ilu_InputCharacter (ilu_Call call, ilu_character *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_character(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfCharacter (ilu_Call call, ilu_character i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_character(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== short integer ==================== */

void
ilu_OutputShortInteger(ilu_Call call, ilu_shortinteger i,
		       ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_short_integer(call_proto(call), call, i, err);
  return;
}

void
ilu_InputShortInteger(ilu_Call call, ilu_shortinteger * i,
		      ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_short_integer(call_proto(call), call, i, err);
  return;
}

/* ARGSUSED */
ilu_cardinal 
ilu_SizeOfShortInteger(ilu_Call call, ilu_shortinteger i,
		       ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_short_integer(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== long integer ==================== */

void ilu_OutputLongInteger (ilu_Call call, ilu_longinteger i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_long_integer(call_proto(call), call, i, err);
  return;
}

void ilu_InputLongInteger (ilu_Call call, ilu_longinteger *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_long_integer(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfLongInteger (ilu_Call call, ilu_longinteger i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_long_integer(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== short cardinal ==================== */

void ilu_OutputShortCardinal (ilu_Call call, ilu_shortcardinal i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_short_cardinal(call_proto(call), call, i, err);
  return;
}

void ilu_InputShortCardinal (ilu_Call call, ilu_shortcardinal *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_short_cardinal(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfShortCardinal (ilu_Call call, ilu_shortcardinal i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_short_cardinal(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== long cardinal ==================== */

void ilu_OutputLongCardinal (ilu_Call call, ilu_longcardinal i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_long_cardinal(call_proto(call), call, i, err);
  return;
}

void ilu_InputLongCardinal (ilu_Call call, ilu_longcardinal *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_long_cardinal(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfLongCardinal (ilu_Call call, ilu_longcardinal i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_long_cardinal(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== enumeration ==================== */

void ilu_OutputEnum (ilu_Call call, ilu_shortcardinal i,
		     ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_enum_code(call_proto(call), call, i, the_type, err);
  return;
}

void ilu_InputEnum (ilu_Call call, ilu_shortcardinal *i,
		    ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_enum_code(call_proto(call), call, i, the_type, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfEnum (ilu_Call call, ilu_shortcardinal i,
			     ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_enum_code(call_proto(call), call, i, the_type, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== real ==================== */

void ilu_OutputReal (ilu_Call call, double i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_real(call_proto(call), call, i, err);
  return;
}

void ilu_InputReal (ilu_Call call, double *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_real(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfReal (ilu_Call call, double i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_real(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== short real ==================== */

void ilu_OutputShortReal (ilu_Call call, float i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_short_real(call_proto(call), call, i, err);
  return;
}

void ilu_InputShortReal (ilu_Call call, float *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_short_real(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfShortReal (ilu_Call call, float i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_short_real(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== long real ==================== */

void ilu_OutputLongReal (ilu_Call call, ilu_longreal i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_long_real(call_proto(call), call, i, err);
  return;
}

void ilu_InputLongReal (ilu_Call call, ilu_longreal *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_long_real(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfLongReal (ilu_Call call, ilu_longreal i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_long_real(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== byte ==================== */

void ilu_OutputByte (ilu_Call call, ilu_byte i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_byte(call_proto(call), call, i, err);
  return;
}

void ilu_InputByte (ilu_Call call, ilu_byte *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_byte(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfByte (ilu_Call call, ilu_byte i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_byte(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== short character ==================== */

void ilu_OutputShortCharacter (ilu_Call call, ilu_shortcharacter i,
			       ILU_ERRS((IoErrs)) * err)
{
  ilu_OutputByte (call, (ilu_byte) i, err);
}

void ilu_InputShortCharacter (ilu_Call call, ilu_shortcharacter *i,
			      ILU_ERRS((IoErrs)) * err)
{
  ilu_InputByte (call, (ilu_byte *) i, err);
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfShortCharacter (ilu_Call call,
				       ilu_shortcharacter i,
				       ILU_ERRS((IoErrs)) * err)
{
  return (ilu_SizeOfByte (call, (ilu_byte) i, err));
}

/* ==================== boolean ==================== */

void ilu_OutputBoolean (ilu_Call call, ilu_boolean i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_boolean(call_proto(call), call, i, err);
  return;
}

void ilu_InputBoolean (ilu_Call call, ilu_boolean *i,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_boolean(call_proto(call), call, i, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfBoolean (ilu_Call call, ilu_boolean i,
		  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal len;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_boolean(call_proto(call), call, i, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== string ==================== */

void
ilu_OutputString(ilu_Call call, ilu_string s, ilu_cardinal len,
		 ilu_cardinal limit,
		 ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq) {
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
    return;
  }
  if (limit > 0 && len > limit) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
#ifdef ENABLE_DEBUGGING
  if (ilu_DebugLevel != 0)
    {
      /* check to see that the string has no octet 0 chars in it,
	 as ILU string types don't allow that */
      if (len > 0 && _ilu_SafeStrlen(s) < len)
	{
	  ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_string_null_char, 0);
	  return;
	}
    }
#endif
  protocol_output_string(call_proto(call), call, s, len, limit, err);
  return;
}

void
ilu_InputString(ilu_Call call, ilu_string * s, ilu_cardinal * len,
		ilu_cardinal limit,
		ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq) {
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
    return;
  }
  if (call_connection(call) == NIL) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
  protocol_input_string(call_proto(call), call, s, len, limit, err);
#ifdef ENABLE_DEBUGGING
  if (ilu_DebugLevel != 0 && ILU_ERROK(*err))
    {
      /* check to see that the string has no octet 0 chars in it,
	 as ILU string types don't allow that */
      if (strlen(*s) < *len)
	ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_string_null_char, 0);
    }
#endif
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfString (ilu_Call call, ilu_string i,
			       ilu_cardinal l, ilu_cardinal limit,
			       ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (limit > 0 && l > limit)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
#ifdef ENABLE_DEBUGGING
  if (ilu_DebugLevel != 0)
    {
      /* check to see that the string has no octet 0 chars in it,
	 as ILU string types don't allow that */
      if (strlen(i) < l)
	{
	  ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_string_null_char, 0);
	  return 0;
	}
    }
#endif
  len = protocol_size_of_string(call_proto(call), call, i, l, limit, err);
  return (ILU_ERROK(*err) ? len : 0);
}

/* ==================== stringvec ==================== */

void 
ilu_OutputStringVec(ilu_Call call, ilu_string i, ilu_cardinal len,
		    ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_stringvec(call_proto(call), call, i, len, err);
  return;
}

void 
ilu_InputStringVec(ilu_Call call, ilu_string * i, ilu_cardinal len,
		   ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_stringvec(call_proto(call), call, i, len, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfStringVec (ilu_Call call, ilu_string i,
				  ilu_cardinal len,
				  ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len2;
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len2 = protocol_size_of_stringvec(call_proto(call), call, i, len,
				    err);
  return (ILU_ERROK(*err) ? len2 : 0);
}

/* ==================== wstring ==================== */

ilu_cardinal _ilu_SizeOfWString (ilu_Call call, ilu_wstring s,
				ilu_cardinal l1	/* size of wstring */,
				ilu_cardinal limit,
				ILU_ERRS((IoErrs)) *err)
{
  ilu_string      buf;
  register ilu_string p1;
  register ilu_character *p2;
  ilu_cardinal    len, size;
  ilu_character  *p2Limit;

  size = protocol_size_of_cardinal(call_proto(call), call, l1, err);
  if (ILU_ERRNOK(*err))
    return 0;
  p2Limit = s + l1;
  buf = ilu_malloc(l1 * 3);
  for (p1 = buf, p2 = s; p2 < p2Limit;) {
    if (*p2 & 0xF800) {
      *p1++ = (char) (0xE0 | (*p2 >> 12));
      *p1++ = (char) (0x80 | (((ilu_cardinal) (*p2 & 0x0FC0)) >> 6));
      *p1++ = (char) (0x80 | (*p2++ & 0x3F));
    } else if (*p2 & 0x0780) {
      *p1++ = (char) (0xC0 | ((ilu_cardinal) (*p2 & 0x07C0)) >> 6);
      *p1++ = (char) (0x80 | (*p2++ & 0x003F));
    } else
      *p1++ = (char) (*p2++ & 0x7F);
  }
  len = p1 - buf;
  size += protocol_size_of_bytes(call_proto(call), call,
				 (ilu_bytes) buf, len, 0, err);
  ilu_free(buf);
  return (ILU_ERROK(*err) ? size : 0);
}

void _ilu_OutputWString (ilu_Call call, ilu_wstring s, ilu_cardinal l1,
			 ilu_cardinal limit, ILU_ERRS((IoErrs)) * err)
{
  ilu_string      buf;
  register ilu_string p1;
  register ilu_character *p2, *p2Limit;
  ilu_cardinal    len;

  protocol_output_cardinal(call_proto(call), call, l1, err);
  if (ILU_ERRNOK(*err))
    return;
  p2Limit = s + l1;
  buf = (ilu_string) ilu_malloc(l1 * 3);
  for (p1 = buf, p2 = s; p2 < p2Limit;) {
    if (*p2 & 0xF800) {
      *p1++ = (char) (0xE0 | (*p2 >> 12));
      *p1++ = (char) (0x80 | (((ilu_cardinal) (*p2 & 0x0FC0)) >> 6));
      *p1++ = (char) (0x80 | (*p2++ & 0x3F));
    } else if (*p2 & 0x0780) {
      *p1++ = (char) (0xC0 | (((ilu_cardinal) (*p2 & 0x07C0)) >> 6));
      *p1++ = (char) (0x80 | (*p2++ & 0x003F));
    } else
      *p1++ = (char) (*p2++ & 0x7F);
  }
  len = p1 - buf;
  protocol_output_bytes(call_proto(call), call, (ilu_bytes) buf,
			len, 0, err);
  ilu_free(buf);
  return;
}

static          ilu_boolean
UTF2Decode(ilu_character * tobuf, ilu_string frombuf,
	   ilu_cardinal tolen, ilu_cardinal fromlen,
	   ILU_ERRS((internal)) * err)
{
  ilu_string      p2 = frombuf, p2lim = frombuf + fromlen;
  ilu_character  *p1 = tobuf;
  ilu_cardinal    count;
  for (count = 0; count < tolen && p2 < p2lim; count++) {
    if ((*p2 & 0xF0) == 0xE0) {
      *p1++ = ((p2[0] & 0x0F) << 12) | ((p2[1] & 0x3F) << 6)
	| (p2[2] & 0x3F);
      p2 += 3;
    } else if ((*p2 & 0xE0) == 0xC0) {
      *p1++ = ((p2[0] & 0x1F) << 6) | (p2[1] & 0x3F);
      p2 += 2;
    } else
      *p1++ = *p2++ & 0x7F;
  }
  if (count == tolen && p2 == p2lim)
    return ILU_CLER(*err);
  else
    return ILU_ERR_CONS1(marshal, err, minor, ilu_mm_utf2Len, FALSE);
}

void
_ilu_InputWString(ilu_Call call, ilu_wstring * s, ilu_cardinal * l,
		  ilu_cardinal limit, ILU_ERRS((IoErrs)) * err)
{
  ilu_string      buf = NIL;
  ilu_character  *ubuf = NIL;
  ilu_cardinal    len = 0, len2;

  protocol_input_cardinal(call_proto(call), call, &len, err);
  if (ILU_ERRNOK(*err))
    return;
  protocol_input_bytes(call_proto(call), call, (ilu_bytes *) & buf,
		       &len2, 0, err);
  if (ILU_ERRNOK(*err))
    return;
  ubuf = (ilu_wstring) ilu_MallocE((len + 1) * sizeof(ilu_character),
				   err);
  if (ubuf == NIL)
    return;
  if (!UTF2Decode(ubuf, buf, len, len2, err)) {
    ilu_free(ubuf);
    *s = NIL;
    *l = 0;
    return;
  }
  ubuf[len] = 0;
  *s = ubuf;
  *l = len;
  return;
}

ilu_cardinal ilu_SizeOfWString (ilu_Call call, ilu_wstring s,
				ilu_cardinal l1	/* size of wstring */,
				ilu_cardinal limit,
				ILU_ERRS((IoErrs)) *err)
{
  ilu_string      buf;
  register ilu_string p1;
  register ilu_character *p2;
  ilu_cardinal    len, size;
  ilu_character  *p2Limit;

  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  if (limit > 0 && l1 > limit)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len = protocol_size_of_wstring(call_proto(call), call, s, l1, limit,
				 err);
  return (ILU_ERROK(*err) ? len : 0);
}

void ilu_OutputWString (ilu_Call call, ilu_wstring s, ilu_cardinal l1,
			ilu_cardinal limit, ILU_ERRS((IoErrs)) * err)
{
  ilu_string      buf;
  register ilu_string p1;
  register ilu_character *p2, *p2Limit;
  ilu_cardinal    len;

  if (call_connection(call) == NIL) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
  if (limit > 0 && l1 > limit) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
  if (call->ca_irq) {
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
    return;
  }
  protocol_output_wstring(call_proto(call), call, s, l1, limit, err);
  return;
}

void
ilu_InputWString(ilu_Call call, ilu_wstring * s, ilu_cardinal * l,
		 ilu_cardinal limit, ILU_ERRS((IoErrs)) * err)
{
  ilu_string      buf = NIL;
  ilu_character  *ubuf = NIL;
  ilu_cardinal    len = 0, len2;

  if (call_connection(call) == NIL) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
  if (call->ca_irq) {
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
    return;
  }
  protocol_input_wstring(call_proto(call), call, s, l, limit, err);
  return;
}

/* ==================== wstringvec ==================== */

void 
  _ilu_OutputWStringVec(ilu_Call call, ilu_wstring i,
			ilu_cardinal len, ILU_ERRS((IoErrs)) * err)
{
  ilu_OutputWString(call, i, len, len, err);
}

void
  _ilu_InputWStringVec(ilu_Call call, ilu_wstring * s,
		       ilu_cardinal len, ILU_ERRS((IoErrs)) * err)
{
  ilu_string      buf = NIL;
  ilu_cardinal    len1 = 0, len2;
  ilu_boolean     isnew = (*s) == NIL;

  protocol_input_cardinal(call_proto(call), call, &len1, err);
  if (ILU_ERRNOK(*err))
    return;
  if (len1 != len) {
    ILU_ERR_CONS1(marshal, err, minor, ilu_mm_wronglen, 6);
    return;
  }
  protocol_input_string(call_proto(call), call, &buf, &len2, 0, err);
  if (ILU_ERRNOK(*err))
    return;
  if (isnew) {
    *s = (ilu_wstring) ilu_MallocE(sizeof(ilu_character) * len + 1,
				   err);
    if (*s == NIL)
      return;
  }
  if (!UTF2Decode(*s, buf, len, len2, err) && isnew) {
    ilu_free(*s);
    *s = NIL;
  }
  return;
}

ilu_cardinal
  _ilu_SizeOfWStringVec (ilu_Call call, ilu_wstring i,
			 ilu_cardinal len, ILU_ERRS((IoErrs)) * err)
{
  return (ilu_SizeOfWString (call, i, len, len, err));
}

void 
ilu_OutputWStringVec(ilu_Call call, ilu_wstring i, ilu_cardinal len,
		     ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_wstringvec(call_proto(call), call, i, len, err);
}

void
ilu_InputWStringVec(ilu_Call call, ilu_wstring * s, ilu_cardinal len,
		    ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_wstringvec(call_proto(call), call, s, len, err);
}

ilu_cardinal ilu_SizeOfWStringVec (ilu_Call call, ilu_wstring i,
				   ilu_cardinal len,
				   ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  return (protocol_size_of_wstringvec(call_proto(call), call, i,
				      len, err));
}

/* ==================== bytes ==================== */

void 
ilu_OutputBytes(ilu_Call call, ilu_bytes i, ilu_cardinal len,
		ilu_cardinal limit, ILU_ERRS((IoErrs)) * err)
{
  if (call_connection(call) == NIL) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
  if (limit > 0 && len > limit)
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_bytes(call_proto(call), call, i, len, limit, err);
  return;
}

void 
ilu_InputBytes(ilu_Call call, ilu_bytes * i, ilu_cardinal * len,
	       ilu_cardinal limit, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_bytes(call_proto(call), call, i, len, limit, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfBytes (ilu_Call call, ilu_bytes i,
			      ilu_cardinal len, ilu_cardinal limit,
			      ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len2;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  if (limit > 0 && len > limit)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len2 = protocol_size_of_bytes(call_proto(call), call, i, len, limit,
				err);
  return (ILU_ERROK(*err) ? len2 : 0);
}

/* ==================== opaque ==================== */

void 
ilu_OutputOpaque(ilu_Call call, ilu_opaque i,
		 ilu_cardinal len, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_opaque(call_proto(call), call, i, len, err);
  return;
}

void 
ilu_InputOpaque(ilu_Call call, ilu_opaque * i,
		ilu_cardinal len, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_opaque(call_proto(call), call, i, len, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfOpaque (ilu_Call call, ilu_opaque i,
			       ilu_cardinal len,
			       ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    len2;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  len2 = protocol_size_of_opaque(call_proto(call), call, i, len, err);
  return (ILU_ERROK(*err) ? len2 : 0);
}

/* ==================== object ==================== */

/**Main Remnant holds.
   obj!=NIL => L1 = {obj's server};
   obj==NIL => L1 = {}*/
ilu_cardinal ilu_SizeOfObjectID(ilu_Call call, ilu_Object h,
				ilu_boolean discriminator_p,
				ilu_Class static_type,
				ILU_ERRS((IoErrs)) *err)
{
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  return protocol_size_of_object_id(call_proto(call), call, h,
				    discriminator_p, static_type,
				    err);
}

/**L2 >= {call's conn's callmu, iomu}.
  h == NIL => Main Invariant holds.
  h != NIL => all the following:
  before: Inside(s, cl);
  after:				 L1 disjoint {cmu, s};
  after: cl collectible		      => L1  not >=  {gcmu};
  after: cl collectible & s surrogate => Main Invariant holds;
  where s = h's server and cl = h's type.
  (We don't really need to hold cmu for surrogate or non-collectible
   objects, but this is convenient because ilu_Enter/ExitServer can
   be used.)*/
void ilu_OutputObjectID (ilu_Call call, ilu_Object h,
			 ilu_boolean discriminator_p,
			 ilu_Class static_type,
			 ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq) {
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
    if (h)
      ilu_ExitServer(h->ob_server, h->ob_class);
  } else
    (void) protocol_output_object_id(call_proto(call), call, h,
				     discriminator_p,
				     static_type, err);
  return;
}

/**Main Remnant holds, Call-IHi(call);
  before: L1 = {},
  after:  *o!=NIL => Inside(*o's server, static_type);
  after:  *o==NIL => L1 = {};
  after:  ILU_ERRNOK(*err) => *o==NIL*/
void ilu_InputObjectID (ilu_Call call, ilu_Object *o,
			ilu_boolean discriminator_p,
			ilu_Class static_type,
			ILU_ERRS((IoErrs)) * err)
{
  *o = NIL;
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    (void) protocol_input_object_id(call_proto(call), call, o,
				    discriminator_p,
				    static_type, err);
  return;
}

/* ==================== generic object ==================== */

/**Call-OHi(call).
   h == NIL => Main Invariant holds.
   h != NIL => all the following:
   before: Inside(s, cl);
   after:				  L1 disjoint {cmu, s};
   after: cl collectible	       => L1  not >=  {gcmu};
   after: cl collectible & s surrogate => Main Invariant holds;
   where s = h's server and cl = h's type.
   (We don't really need to hold cmu for surrogate or non-collectible
    objects, but this is convenient because ilu_Enter/ExitServer can
    be used.)*/
ilu_boolean 
_ilu_OutputObjectID(ilu_Call call, ilu_Object h,
		    ilu_boolean discriminator_p,
		    ilu_Class static_type,
		    ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal	server_id_hash;
  ilu_string      ostr2 = NIL;
  ilu_boolean     is_nil = (h == NIL);

  if (discriminator_p) {
    if (h == NIL)
      return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
    server_id_hash = object_server(h)->sr_crc32;
    ostr2 = h->ob_ih;
  } else if (h == NIL) {
    if (!static_type->cl_optional)
      return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
    ostr2 = "";
  } else {
    ostr2 = ilu_SBHOfObject(h);
  }

  if (ostr2 == NIL) {
    ilu_ExitServer(object_server(h), object_class(h));
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_not_exported,
			 FALSE);
  }
  if (!is_nil) {
    if (object_is_true(h) && object_collectible(h)) {
      ILU_ERRS((BadDataStructure, KernelBroken)) lerr;
      object_lastRemote(h) = ilu_CoarseTime_Now();
      lerr = _ilu_TouchedObj(h);
      ILU_MUST_BE_SUCCESS(lerr);
    }
    ilu_ExitServer(object_server(h), object_class(h));
  }
  if (discriminator_p) {
    protocol_output_cardinal(call_proto(call), call, server_id_hash, err);
    if (ILU_ERRNOK(*err))
      return FALSE;
  }
  protocol_output_string(call_proto(call), call, ostr2,
			 _ilu_SafeStrlen(ostr2), 0xFFFF, err);
  return (ILU_ERROK(*err));
}

/**before: L1 = {};
   after:  *h!=NIL => Inside(*h's server, static_type);
   after:  *h==NIL => L1 = {};
   Main Remnant holds, Call-IHi(call)*/
ilu_boolean 
_ilu_InputObjectID(ilu_Call call, ilu_Object * h,
		   ilu_boolean discriminator_p,
		   ilu_Class static_type,
		   ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal server_id_hash;
  ilu_string      istr2 = NIL;
  ilu_cardinal    istrlen2 = 0;
  ilu_Server      server = connection_server(call_connection(call));

  *h = NIL;
  if (static_type == NIL) {
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, FALSE);
  }
  if (discriminator_p) {
    protocol_input_cardinal (call_proto(call),
			     call, &server_id_hash, err);
    if (ILU_ERRNOK(*err))
      return FALSE;
    ILU_NOTE(INCOMING_DEBUG | OBJECT_DEBUG,
	  ("ilu_InputObjectID:  server id hash is %lx\n",
	   server_id_hash));
  }
  protocol_input_string(call_proto(call),
			call, &istr2, &istrlen2, 0xFFFF, err);
  if (ILU_ERRNOK(*err)) {
    return FALSE;
  }
  ILU_NOTE(INCOMING_DEBUG,
	("ilu_InputObjectID:  instance handle/sbh is <%s>\n",
	 istr2));
  ILU_CLER(*err);
  if (discriminator_p) {
    ilu_string      ih = istr2;
    ilu_EnterServer(server, static_type);
    if (server_id_hash != server->sr_crc32) {
      ILU_NOTE(INCOMING_DEBUG,
	    ("%s %lx is for wrong server (not expected %lx, for <%s>).\n",
	     "ilu_InputObjectID:  incoming sid hash", server_id_hash,
	     server->sr_crc32, server_id(server)));
      (void) ILU_ERR_CONS1(marshal, err, minor, ilu_mm_alien_disc, 6);
      ilu_ExitServer(server, static_type);
    } else if (server_objs(server) == NIL) {
      ILU_NOTE(INCOMING_DEBUG,
	    ("%s %s is in closed server <%s>.\n",
	     "ilu_InputObjectID:  instance", ih, server_id(server)));
      (void) ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_svr_closed, 6);
      ilu_ExitServer(server, static_type);
    } else if ((*h = _ilu_FindObjectInServer(ih, server))
	       == NIL) {
      ILU_NOTE(INCOMING_DEBUG,
	    ("%s %s not found in server <%s>.\n",
	     "ilu_InputObjectID:  instance", ih, server_id(server)));
      (void) ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_inst_nf, 6);
      ilu_ExitServer(server, static_type);
    } else if (!ilu_IsSubObjectType((*h)->ob_class, static_type)) {
      ILU_NOTE(INCOMING_DEBUG,
	    ("%s %s/%s has type %s (%s), not a subtype of %s (%s).\n",
	     "_ilu_InputObjectID: Existing object",
	     server_id(server), ih,
	     (*h)->ob_class->cl_unique_id, (*h)->ob_class->cl_name,
	     static_type->cl_unique_id, static_type->cl_name));
      (void) ILU_ERR_CONS1(inv_objref, err, minor, ilu_iom_wrong_type, 6);
      *h = NIL;
      ilu_ExitServer(server, static_type);
    }
    if (ILU_ERRNOK(*err)) {
      FREETOKEN(istr2);
      return (FALSE);
    }
  } else {
    if (istrlen2 == 0) {
      *h = NIL;
      if (static_type->cl_optional) {
	if (istr2 != NIL)
	  ilu_free(istr2);
	return TRUE;
      } else {
	ILU_NOTE(INCOMING_DEBUG,
	      ("ilu_InputObjectID:  bad NIL obj.\n"));
	return ILU_ERR_CONS1(NoObjectForSBH, err, sbh, istr2, FALSE);
      }
    } else {
      *h = ilu_ObjectOfSBH(istr2, static_type, err);
      if (ILU_ERRNOK(*err)) {
	ILU_NOTE(INCOMING_DEBUG,
	      ("ilu_InputObjectID:  error:  No object for SBH <%s>.\n",
	       istr2));
	ilu_free(istr2);
	return FALSE;
      }
    }
  }

  ilu_free(istr2);
  return (TRUE);
}

/*h!=NIL => L1 >= {h's server}, L1_sup < prmu;
  h==NIL => L1 unconstrained*/
ilu_cardinal _ilu_SizeOfObjectID(ilu_Call call, ilu_Object h,
				 ilu_boolean discriminator_p,
				 ilu_Class static_type,
				 ILU_ERRS((IoErrs)) *err)
{
  ilu_cardinal	server_id_hash;
  ilu_string      strB;		/* discriminator_p ? ih : SBH */
  ilu_cardinal    size1, size2;
  ilu_Connection  conn = call_connection(call);

  ILU_CLER(*err);		/* assume success */

  if (conn == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);

  if (h == NIL) {
    strB = "";
  } else if (discriminator_p) {
    server_id_hash = h->ob_server->sr_crc32;
    strB = h->ob_ih;
  } else {
    strB = ilu_SBHOfObject(h);
  }
  if (strB == NIL)
    return ILU_ERR_CONS1(internal, err, minor, ilu_im_broken, 0);

  if (discriminator_p) {
    size1 = protocol_size_of_cardinal(connection_protocol(conn), call,
				      server_id_hash, err);
    if (ILU_ERRNOK(*err))
      return 0;
  } else
    size1 = 0;

  size2 = protocol_size_of_string(connection_protocol(conn), call,
				  strB, _ilu_SafeStrlen(strB),
				  0xFFFF, err);
  if (ILU_ERRNOK(*err))
    return 0;
  return (size1 + size2);
}

/**before: Call-Locking(call, IHi);
    after: Call-Remnant(call, err, TRUE);
    after: call->ca_ms==ilu_cmsHi && call->ca_msInput;
    after: result!=NIL => Inside(call->ca_server, call->ca_intro_type);
    after: result==NIL => L1 = {}*/
ilu_Object 
ilu_GetCallSingleton(ilu_Call call,
		     ILU_ERRS((bad_param)) * err)
{
  ilu_Server      s = call->ca_server;
  ilu_Class       intro_type = call->ca_intro_type;
  ilu_Object      ans = NIL;
  ilu_EnterServer(s, intro_type);
  if (server_singles(s) != NIL)
    ans = (ilu_Object) ilu_hash_FindInTable(server_singles(s),
					     intro_type);
  if (ans != NIL)
    return (ILU_CLER(*err), ans);
  ilu_ExitServer(s, intro_type);
  call->ca_pe = ilu_ProtocolException_NoSuchClassAtServer;
  return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, NIL);
}

/*************************************************************
**************************************************************
**************************************************************

  Code for sequences, records, arrays, and unions

**************************************************************
**************************************************************
*************************************************************/

void 
ilu_OutputSequence(ilu_Call call, ilu_cardinal i, ilu_cardinal limit,
		   ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq) {
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
    return;
  }
  if (limit > 0 && i > limit) {
    ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
    return;
  }
  protocol_output_sequence(call_proto(call), call, i, limit, the_type, err);
  return;
}

void ilu_OutputSequenceMark (ilu_Call call, ilu_cardinal extent,
		  ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_sequence_mark(call_proto(call), call, extent, err);
  return;
}

void
ilu_InputSequenceMark(ilu_Call call, ilu_cardinal extent,
		      ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_sequence_mark(call_proto(call), call,
				      extent, err);
  return;
}

void 
ilu_InputSequence(ilu_Call call, ilu_cardinal * i,
		  ilu_cardinal limit, ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_sequence(call_proto(call), call, i, limit, the_type, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal
ilu_SizeOfSequence(ilu_Call call, ilu_cardinal i, ilu_cardinal limit,
		   ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    size;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  if (limit > 0 && i > limit)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  size = protocol_size_of_sequence(call_proto(call), call,
				   i, limit, the_type, err);
  return (ILU_ERROK(*err) ? size : 0);
}

/*L1, L2 unconstrained*/
ilu_boolean ilu_EndSequence (ilu_Call call, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_end_sequence(call_proto(call), call, err);
  return ILU_ERROK(*err);
}

void 
ilu_OutputUnion(ilu_Call call, ilu_cardinal i,
		ilu_TypeKind typekind,
		ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_union(call_proto(call), call, i, typekind, the_type, err);
  return;
}

void 
ilu_InputUnion(ilu_Call call, ilu_cardinal * i, ilu_TypeKind typekind,
	       ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_union(call_proto(call), call, i, typekind, the_type, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal 
ilu_SizeOfUnion(ilu_Call call, ilu_cardinal i, ilu_TypeKind typekind,
		ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    size;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  size = protocol_size_of_union(call_proto(call), call, i, typekind, the_type, err);
  return (ILU_ERROK(*err) ? size : 0);
}

/*L1, L2 unconstrained*/
ilu_boolean ilu_EndUnion (ilu_Call call, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_end_union(call_proto(call), call, err);
  return ILU_ERROK(*err);
}

void 
ilu_OutputArray(ilu_Call call, ilu_cardinal length,
		ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_array(call_proto(call), call, length, the_type, err);
  return;
}

void 
ilu_InputArray(ilu_Call call, ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_array(call_proto(call), call, the_type, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal 
ilu_SizeOfArray(ilu_Call call, ilu_cardinal length,
		ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    size;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  size = protocol_size_of_array(call_proto(call), call,
				length, the_type, err);
  return (ILU_ERROK(*err) ? size : 0);
}

/*L1, L2 unconstrained*/
ilu_boolean ilu_EndArray (ilu_Call call, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_end_array(call_proto(call), call, err);
  return ILU_ERROK(*err);
}

void 
ilu_OutputRecord(ilu_Call call, ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_output_record(call_proto(call), call, the_type, err);
  return;
}

void 
ilu_InputRecord(ilu_Call call, ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_input_record(call_proto(call), call, the_type, err);
  return;
}

/*ARGSUSED*/
ilu_cardinal ilu_SizeOfRecord (ilu_Call call, ilu_Type the_type, ILU_ERRS((IoErrs)) * err)
{
  ilu_cardinal    size;
  if (call_connection(call) == NIL)
    return ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_duh, 0);
  else if (call->ca_irq || !protocol_needs_sizing(call_proto(call)))
    return (ILU_CLER(*err), 0);
  size = protocol_size_of_record(call_proto(call), call, the_type, err);
  return (ILU_ERROK(*err) ? size : 0);
}

/*L1, L2 unconstrained*/
ilu_boolean ilu_EndRecord (ilu_Call call, ILU_ERRS((IoErrs)) *err)
{
  if (call->ca_irq)
    ILU_ERR_CONS1(interrupted, err, ilu_interruptSet, 0, (void) 6);
  else
    protocol_end_record(call_proto(call), call, err);
  return ILU_ERROK(*err);
}

/*L1, L2 unconstrained*/

ilu_Passport ilu_CallerPassportOfCall (ilu_Call call)
{
  return (call->ca_caller);
}

void ilu_SetCallerPassportOfCall (ilu_Call call, ilu_Passport pport)
{
  call->ca_caller = pport;
}
